import logging
from dataclasses import dataclass
from typing import Iterable

from ofrak.core.addressable import Addressable
from ofrak.model.resource_model import index, ResourceAttributes
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak_type.error import NotFoundError
from ofrak_type.range import Range

LOGGER = logging.getLogger(__file__)


@dataclass
class MemoryRegion(Addressable):
    """
    Binary bytes that are addressable.

    :ivar virtual_address: the virtual address of the start of the memory region
    :ivar size: the size of the memory region
    """

    size: int

    @index
    def Size(self) -> int:
        return self.size

    @index(uses_indexes=(Addressable.VirtualAddress,))
    def EndVaddr(self) -> int:
        return self.size + self.VirtualAddress

    def end_vaddr(self) -> int:
        """
        Get the virtual address of the end of the memory region.

        :returns: the virtual address directly after the memory region
        """
        return self.virtual_address + self.size

    def vaddr_range(self) -> Range:
        return Range.from_size(self.virtual_address, self.size)

    @classmethod
    def caption(cls, all_attributes) -> str:
        try:
            mem_region_attributes = all_attributes[AttributesType[MemoryRegion]]
            addressable_attributes = all_attributes[AttributesType[Addressable]]
        except KeyError:
            return super().caption(all_attributes)
        return (
            f"{str(cls.__name__)}: "
            f"{hex(addressable_attributes.virtual_address)}-"
            f"{hex(addressable_attributes.virtual_address + mem_region_attributes.size)}"
        )

    def contains(self, vaddr: int) -> bool:
        """
        Does the memory region contain the given virtual address?

        :param vaddr: a virtual address

        :return: True if the memory region contains the given virtual address
        """
        return self.virtual_address <= vaddr < self.end_vaddr()

    def get_offset_in_self(self, vaddr: int) -> int:
        """
        Get the physical offset within the memory region that corresponds to the given virtual
        address.

        :param vaddr: a virtual address

        :return: an offset within the memory region
        """
        if not self.contains(vaddr):
            raise ValueError(
                f"Memory region {hex(self.virtual_address)}-{hex(self.end_vaddr())} "
                f"does not contain vaddr {hex(vaddr)}"
            )
        return vaddr - self.virtual_address

    async def create_child_region(
        self,
        child_mr: "MemoryRegion",
        additional_attributes: Iterable[ResourceAttributes] = (),
    ) -> Resource:
        """
        Create a child memory region that is mapped into this memory region.

        :param child_mr: the child memory region
        :param additional_attributes: additional attributes passed to the child memory region

        :raises ValueError: if the child's end offset is larger than the memory region's size
        :return: the created child resource
        """
        start_offset = self.get_offset_in_self(child_mr.virtual_address)
        end_offset = start_offset + child_mr.size
        if start_offset < 0:
            raise ValueError(
                f"New child has vaddr {hex(child_mr.virtual_address)} which is before"
                f" the proposed parent's vaddr {hex(self.virtual_address)}"
            )
        if end_offset > self.size:
            raise ValueError(
                f"New child at {hex(child_mr.virtual_address)} is too large to fit in the proposed "
                f"parent - end vaddr {hex(child_mr.end_vaddr())} goes past the parent's end vaddr "
                f"{hex(self.end_vaddr())}."
            )

        return await self.resource.create_child_from_view(
            child_mr,
            data_range=Range(start_offset, end_offset),
            additional_attributes=additional_attributes,
        )

    @staticmethod
    def get_mem_region_with_vaddr_from_sorted(vaddr: int, sorted_regions: Iterable["MemoryRegion"]):
        """
        Return the first [memory region][ofrak.core.memory_region.MemoryRegion] in the input
        iterable that contains vaddr.

        :param vaddr: Virtual address
        :param sorted_regions: Sorted iterable of memory regions to check in order for vaddr

        :raises NotFoundError: If vaddr is not in any element of the iterable

        :return: The first memory region in the sorted iterable containing vaddr
        """
        for mem_view in sorted_regions:
            # the first region we find should be the largest
            mem_region_vaddr_range = Range(
                mem_view.virtual_address,
                mem_view.virtual_address + mem_view.size,
            )
            if vaddr in mem_region_vaddr_range:
                return mem_view

        raise NotFoundError(f"Cannot find memory region matching {hex(vaddr)}")

    def __str__(self):
        if type(self) is MemoryRegion:
            return f"MemoryRegion({hex(self.virtual_address)}-{hex(self.end_vaddr())})"
        else:
            return super().__str__()

    def __hash__(self):
        """
        Return the hash of the virtual address and size.

        !!! warning

            Two memory regions may have the same hash, even if they refer to different data! As
            long as the address and size are the same, two regions will have the same hash,
            since the resource is not part of the data that is hashed. Be careful about comparing
            memory regions that refer to different data!
        """
        return hash((self.virtual_address, self.size))



import logging
import os
from dataclasses import dataclass
from io import BytesIO
from typing import Iterable, Optional

from pycdlib import PyCdlib

from ofrak.component.analyzer import Analyzer
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import FilesystemRoot, File, Folder
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.resource_model import index
from ofrak.resource import Resource
from ofrak.resource_view import ResourceView
from ofrak.service.resource_service_i import ResourceFilter, ResourceAttributeValueFilter
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class ISO9660ImageAttributes(ResourceAttributes):
    interchange_level: int
    volume_identifier: str
    system_identifier: str
    app_identifier: str
    extended_attributes: bool
    has_joliet: bool
    has_rockridge: bool
    has_udf: bool
    has_eltorito: bool
    joliet_level: Optional[int] = None
    rockridge_version: Optional[str] = None
    udf_version: Optional[str] = None


@dataclass
class ISO9660Entry(ResourceView):
    name: str
    path: str
    is_dir: bool
    is_file: bool
    is_symlink: bool
    is_dot: bool
    is_dotdot: bool
    iso_version: int

    @index
    def Path(self) -> str:
        return self.path


@dataclass
class ISO9660Image(GenericBinary, FilesystemRoot):
    """
    ISO 9660 image. ISO 9660 is a file system for optical disc media.
    """

    async def get_file(self, path: str) -> ResourceView:
        return await self.resource.get_only_descendant_as_view(
            ISO9660Entry,
            r_filter=ResourceFilter(
                attribute_filters=(ResourceAttributeValueFilter(ISO9660Entry.Path, path),)
            ),
        )

    async def get_entries(self) -> Iterable[ISO9660Entry]:
        return await self.resource.get_descendants_as_view(
            ISO9660Entry, r_filter=ResourceFilter(tags=(ISO9660Entry,))
        )


@dataclass
class JolietISO9660Image(ISO9660Image):
    pass


@dataclass
class UdfISO9660Image(ISO9660Image):
    pass


@dataclass
class RockRidgeISO9660Image(ISO9660Image):
    pass


@dataclass
class ElToritoISO9660Image(ISO9660Image):
    pass


class ISO9660ImageAnalyzer(Analyzer[None, ISO9660ImageAttributes]):
    targets = (ISO9660Image,)
    outputs = (ISO9660ImageAttributes,)

    async def analyze(self, resource: Resource, config=None):
        joliet_level = None
        rockridge_version = None
        udf_version = None

        iso = PyCdlib()
        iso.open_fp(BytesIO(await resource.get_data()))

        interchange_level = iso.interchange_level
        has_joliet = iso.has_joliet()
        has_rockridge = iso.has_rock_ridge()
        has_udf = iso.has_udf()
        has_eltorito = iso.eltorito_boot_catalog is not None
        vol_identifier = iso.pvd.volume_identifier.decode("utf-8").strip()
        sys_identifier = iso.pvd.system_identifier.decode("utf-8").strip()
        app_identifier = iso.pvd.application_identifier.text.decode("utf-8").strip()
        xa = iso.xa

        if has_joliet:
            esc = iso.joliet_vd.escape_sequences
            if b"%/@" in esc:
                joliet_level = 1
            elif b"%/C" in esc:
                joliet_level = 2
            elif b"%/E" in esc:
                joliet_level = 3
        elif has_rockridge:
            rockridge_version = iso.rock_ridge
        elif has_udf:
            udf_version = "2.60"

        iso.close()

        return ISO9660ImageAttributes(
            interchange_level=interchange_level,
            volume_identifier=vol_identifier,
            system_identifier=sys_identifier,
            app_identifier=app_identifier,
            extended_attributes=xa,
            has_joliet=has_joliet,
            joliet_level=joliet_level,
            has_rockridge=has_rockridge,
            rockridge_version=rockridge_version,
            has_udf=has_udf,
            udf_version=udf_version,
            has_eltorito=has_eltorito,
        )


class ISO9660Unpacker(Unpacker[None]):
    """
    Unpack an ISO 9660 image.
    """

    id = b"ISO9660Unpacker"
    targets = (ISO9660Image,)
    children = (ISO9660Entry,)

    async def unpack(self, resource: Resource, config=None):
        iso_data = await resource.get_data()

        iso_attributes = await resource.analyze(ISO9660ImageAttributes)
        resource.add_attributes(iso_attributes)
        iso_resource = await resource.view_as(ISO9660Image)

        iso = PyCdlib()
        iso.open_fp(BytesIO(iso_data))

        if iso_attributes.has_joliet:
            facade = iso.get_joliet_facade()
            path_var = "joliet_path"
        elif iso_attributes.has_udf:
            LOGGER.warning("UDF images are not currently supported")
            facade = iso.get_udf_facade()
            path_var = "udf_path"
        elif iso_attributes.has_rockridge:
            LOGGER.warning("Rock Ridge images are not currently supported")
            facade = iso.get_rock_ridge_facade()
            path_var = "rr_name"
        else:
            facade = iso.get_iso9660_facade()
            path_var = "iso_path"

        if iso_attributes.has_eltorito:
            LOGGER.warning("El Torito images are not currently supported")

        for root, dirs, files in iso.walk(**{path_var: "/"}):
            for d in dirs:
                path = os.path.join(root, d)
                folder_tags = (ISO9660Entry, Folder)
                entry = ISO9660Entry(
                    name=d,
                    path=path,
                    is_dir=True,
                    is_file=False,
                    is_symlink=False,
                    is_dot=(str(d).startswith(".") and not str(d).startswith("..")),
                    is_dotdot=str(d).startswith(".."),
                    iso_version=-1,
                )
                await iso_resource.add_folder(
                    path, None, None, folder_tags, entry.get_attributes_instances().values()
                )
            for f in files:
                path = os.path.join(root, f)
                file_tags = (ISO9660Entry, File)
                fp = BytesIO()

                facade.get_file_from_iso_fp(fp, **{path_var: path})
                file_data = fp.getvalue()

                if ";" in f:
                    f, iso_version = f.split(";")
                    iso_version = int(iso_version)
                    path = path.split(";")[0]

                    if f.endswith("."):
                        f = f[:-1]
                        path = path[:-1]
                else:
                    iso_version = -1

                entry = ISO9660Entry(
                    name=f,
                    path=path,
                    is_dir=False,
                    is_file=True,
                    is_symlink=False,
                    is_dot=(str(f).startswith(".") and not str(f).startswith("..")),
                    is_dotdot=str(f).startswith(".."),
                    iso_version=iso_version,
                )
                await iso_resource.add_file(
                    path,
                    file_data,
                    None,
                    None,
                    file_tags,
                    entry.get_attributes_instances().values(),
                )
                fp.close()

        iso.close()


class ISO9660Packer(Packer[None]):
    """
    Pack files in an ISO 9660 image.
    """

    id = b"ISO9660Packer"
    targets = (ISO9660Image,)

    async def pack(self, resource: Resource, config=None):
        image_attributes = resource.get_attributes(ISO9660ImageAttributes)

        iso_result = PyCdlib()
        iso_result.new(
            interchange_level=image_attributes.interchange_level,
            sys_ident=image_attributes.system_identifier,
            vol_ident=image_attributes.volume_identifier,
            app_ident_str=image_attributes.app_identifier,
            joliet=image_attributes.joliet_level,
            rock_ridge=image_attributes.rockridge_version,
            xa=image_attributes.extended_attributes,
            udf=image_attributes.udf_version,
        )

        if image_attributes.has_joliet:
            resource.add_tag(JolietISO9660Image)
            facade = iso_result.get_joliet_facade()
            path_arg = "joliet_path"
        elif image_attributes.has_udf:
            resource.add_tag(UdfISO9660Image)
            facade = iso_result.get_udf_facade()
            path_arg = "udf_path"
            LOGGER.warning("UDF images are not currently supported")
        elif image_attributes.has_rockridge:
            LOGGER.warning("Rock Ridge images are not currently supported")
            resource.add_tag(RockRidgeISO9660Image)
            facade = iso_result.get_rock_ridge_facade()
            path_arg = "rr_name"
        else:
            facade = iso_result.get_iso9660_facade()
            path_arg = "iso_path"

        if image_attributes.has_eltorito:
            LOGGER.warning("El Torito images are not currently supported")

        child_queue = [
            d
            for d in await resource.get_children_as_view(
                ISO9660Entry, r_filter=ResourceFilter(tags=(ISO9660Entry,))
            )
        ]

        while len(child_queue) > 0:
            child = child_queue.pop(0)

            path = child.path
            if child.iso_version != -1:
                path += ";" + str(child.iso_version)

            if child.resource.has_tag(Folder):
                facade.add_directory(**{path_arg: path})
                for d in await child.resource.get_children_as_view(
                    ISO9660Entry, r_filter=ResourceFilter(tags=(ISO9660Entry,))
                ):
                    child_queue.append(d)
            elif child.resource.has_tag(File):
                file_data = await child.resource.get_data()
                facade.add_fp(BytesIO(file_data), len(file_data), **{path_arg: path})

        result = BytesIO()
        iso_result.write_fp(result)
        iso_result.close()

        iso_data = result.getvalue()
        result.close()
        resource.queue_patch(Range(0, await resource.get_data_length()), iso_data)


MagicMimeIdentifier.register(ISO9660Image, "application/x-iso9660-image")
MagicDescriptionIdentifier.register(ISO9660Image, lambda s: s.startswith("ISO 9660 CD"))



import heapq
import itertools
from bisect import bisect_left, bisect_right
from collections import defaultdict
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Generic,
    Pattern,
    cast,
)

from sortedcontainers import SortedList

from ofrak.model.data_model import DataModel, DataPatch, DataPatchesResult
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.error import OutOfBoundError, PatchOverlapError
from ofrak_type.error import NotFoundError, AlreadyExistError
from ofrak_type.range import Range

# Type alias; typechecker makes no distinction between this and bytes. It's just for humans (you?)
DataId = bytes


class DataService(DataServiceInterface):
    def __init__(self):
        self._model_store: Dict[DataId, DataModel] = dict()
        self._roots: Dict[DataId, _DataRoot] = dict()

    async def create_root(self, data_id: DataId, data: bytes) -> DataModel:
        if data_id in self._model_store:
            raise AlreadyExistError(f"A model with {data_id.hex()} already exists!")

        new_model = DataModel(data_id, Range(0, len(data)), data_id)

        self._model_store[data_id] = new_model
        self._roots[data_id] = _DataRoot(new_model, data)

        return new_model

    async def create_mapped(
        self,
        data_id: DataId,
        parent_id: DataId,
        range_in_parent: Range,
    ) -> DataModel:
        if data_id in self._model_store:
            raise AlreadyExistError(f"A model with {data_id.hex()} already exists!")

        parent_model = self._get_by_id(parent_id)
        range_in_root = range_in_parent.translate(parent_model.range.start)
        if range_in_root.end > parent_model.range.end:
            raise OutOfBoundError(
                f"Cannot map a new node into range {range_in_root} into {parent_model.range} of "
                f"{parent_id.hex()}"
            )

        new_model = DataModel(data_id, range_in_root, parent_model.root_id)
        self._roots[parent_model.root_id].add_mapped_model(new_model)
        self._model_store[data_id] = new_model

        return new_model

    async def get_by_id(self, data_id: DataId) -> DataModel:
        return self._get_by_id(data_id)

    async def get_by_ids(self, data_ids: Iterable[DataId]) -> Iterable[DataModel]:
        return [self._get_by_id(data_id) for data_id in data_ids]

    async def get_data_length(self, data_id: DataId) -> int:
        return self._get_by_id(data_id).range.length()

    async def get_data_range_within_root(self, data_id: DataId) -> Range:
        return self._get_by_id(data_id).range

    async def get_range_within_other(self, data_id: DataId, within_data_id: DataId) -> Range:
        model = self._get_by_id(data_id)
        within_model = self._get_by_id(within_data_id)
        if data_id == within_data_id:
            return Range.from_size(0, model.range.length())
        if self._is_root(data_id):
            raise ValueError(
                f"{data_id.hex()} is a root, not mapped into {within_data_id.hex()} (a root)!"
            )
        elif self._is_root(within_data_id) and model.root_id != within_model.id:
            raise ValueError(f"{data_id.hex()} is not mapped into {within_data_id.hex()} (a root)!")
        elif not self._is_root(within_data_id) and model.root_id != within_model.root_id:
            raise ValueError(
                f"{data_id.hex()} and {within_data_id.hex()} are not mapped into the same root!"
            )
        else:
            return within_model.range.intersect(model.range).translate(-within_model.range.start)

    async def get_data(self, data_id: DataId, data_range: Optional[Range] = None) -> bytes:
        model = self._get_by_id(data_id)
        root = self._get_root_by_id(model.root_id)
        if data_range is not None:
            translated_range = data_range.translate(model.range.start).intersect(root.model.range)
            return root.data[translated_range.start : translated_range.end]
        else:
            return root.data[model.range.start : model.range.end]

    async def apply_patches(self, patches: List[DataPatch]) -> List[DataPatchesResult]:
        patches_by_root: Dict[DataId, List[DataPatch]] = defaultdict(list)
        for patch in patches:
            target_data_model = self._get_by_id(patch.data_id)
            patches_by_root[target_data_model.root_id].append(patch)

        results = []
        for root_id, patches_for_root in patches_by_root.items():
            results.extend(self._apply_patches_to_root(root_id, patches_for_root))

        return results

    async def delete_models(self, data_ids: Iterable[DataId]) -> None:
        roots_to_delete = dict()
        mapped_to_delete = dict()

        for data_id in data_ids:
            try:
                model = self._get_by_id(data_id)
            except NotFoundError:
                continue
            if model.is_mapped():
                mapped_to_delete[model.id] = model
            else:
                roots_to_delete[model.id] = model

        for root_model in roots_to_delete.values():
            root = self._roots[root_model.id]
            for child_model in root.get_children():
                mapped_to_delete.pop(child_model.id, None)
                del self._model_store[child_model.id]

            del self._roots[root_model.id]
            del self._model_store[root_model.id]

        for model in mapped_to_delete.values():
            root = self._get_root_by_id(model.root_id)
            root.delete_mapped_model(model)
            del self._model_store[model.id]

    async def search(self, data_id, query, start=None, end=None, max_matches=None):
        model = self._get_by_id(data_id)
        root = self._get_root_by_id(model.root_id)
        start = model.range.start if start is None else model.range.start + start
        end = model.range.end if end is None else min(model.range.end, model.range.start + end)
        if isinstance(query, bytes):
            matches = []
            while max_matches is None or len(matches) < max_matches:
                match_offset = root.data.find(query, start, end)
                if match_offset < 0:
                    break

                matches.append(match_offset - model.range.start)
                start = match_offset + 1

            return tuple(matches)
        else:
            query = cast(Pattern, query)
            match_iterator = query.finditer(root.data, start, end)

            if max_matches is not None:
                match_iterator = itertools.islice(match_iterator, max_matches)
            matches = (
                (match.start() - model.range.start, match.group(0)) for match in match_iterator
            )
            return tuple(matches)

    def _get_by_id(self, data_id: DataId) -> DataModel:
        model = self._model_store.get(data_id)
        if model is None:
            raise NotFoundError(f"No data model with ID {data_id.hex()} exists")
        else:
            return model

    def _get_root_by_id(self, data_id: DataId) -> "_DataRoot":
        root = self._roots.get(data_id)
        if root is None:
            raise NotFoundError(f"No data root with ID {data_id.hex()} exists")
        else:
            return root

    def _is_root(self, data_id: DataId) -> bool:
        return data_id in self._roots

    def _get_range_in_root(self, model: DataModel, r: Range) -> Range:
        if r.start < 0 or r.end > model.range.length():
            raise OutOfBoundError(
                f"The requested range {r} of model {model.id.hex()} is outside the "
                f"model's range {model.range}"
            )

        absolute_range = r.translate(model.range.start)
        return absolute_range

    def _apply_patches_to_root(
        self,
        root_data_id: DataId,
        patches: List[DataPatch],
    ) -> List[DataPatchesResult]:
        root: _DataRoot = self._roots[root_data_id]
        finalized_ordered_patches: List[Tuple[Range, bytes, int]] = []
        resize_tracker = _PatchResizeTracker()

        # Screen patches for inconsistencies/overlaps
        # And translate them as required by preceding patches
        raw_patch_ranges_in_root = []

        for patch in patches:
            target_model = self._get_by_id(patch.data_id)
            patch_range_in_prepatch_root = self._get_range_in_root(target_model, patch.range)

            if resize_tracker.overlaps_resized_range(patch_range_in_prepatch_root):
                raise PatchOverlapError(
                    f"Patch to {patch.range} of {patch.data_id.hex()} overlaps previously resized "
                    f"area of {patch.data_id.hex()} and cannot be applied!"
                )

            raw_patch_ranges_in_root.append(patch_range_in_prepatch_root)
            patch_range_in_patched_root = resize_tracker.translate_range(
                patch_range_in_prepatch_root
            )
            models_intersecting_patch_range = root.get_children_with_boundaries_intersecting_range(
                patch_range_in_prepatch_root
            )

            size_diff = len(patch.data) - patch.range.length()
            if size_diff == 0:
                finalized_ordered_patches.append(
                    (patch_range_in_patched_root, patch.data, size_diff)
                )
            elif models_intersecting_patch_range:
                raise PatchOverlapError(
                    f"Because patch to {patch.data_id.hex()} resizes data by {size_diff} bytes, "
                    f"the effects on {len(models_intersecting_patch_range)} model(s) "
                    f"intersecting the patch range {patch.range} ({patch_range_in_patched_root} "
                    f"in the root) could not be determined. If data must be resized, any resources "
                    f"overlapping the data must be deleted before patching and re-created "
                    f"afterwards along new data ranges. Intersecting models: {models_intersecting_patch_range}"
                )
            else:
                finalized_ordered_patches.append(
                    (patch_range_in_patched_root, patch.data, size_diff)
                )
                resize_tracker.add_new_resized_range(patch_range_in_patched_root, size_diff)

        affected_ranges = Range.merge_ranges(raw_patch_ranges_in_root)

        results = defaultdict(list)

        for affected_id, affected_range in root.get_children_affected_by_ranges(affected_ranges):
            results[affected_id].append(affected_range)

        for affected_range in affected_ranges:
            results[root_data_id].append(affected_range)

        new_root_data = bytearray(root.data)
        # Apply finalized patches to data and data models
        for patch_range, data, size_diff in finalized_ordered_patches:
            new_root_data[patch_range.start : patch_range.end] = data
            if size_diff != 0:
                root.resize_range(patch_range, size_diff)
        root.data = bytes(new_root_data)

        return [
            DataPatchesResult(data_id, results_for_id)
            for data_id, results_for_id in results.items()
        ]


T = TypeVar("T")


class _ShiftBreaksSortError(RuntimeError):
    pass


class _CompareFirstTuple(Tuple, Generic[T]):
    """
    Wrapper for tuple that ensures only the first item in the tuple is checked.
    Necessary because bisect methods don't have a `key` function
    Helpful for making sorted dictionary-like data structures
    """

    def __new__(cls, *args):
        return super().__new__(cls, args)

    def __lt__(self, other):
        return self[0] < other[0]

    def __eq__(self, other):
        return self[0] == other[0]

    # __gt__ function excluded, as it is not needed by built-in Python utils like `sort`

    @staticmethod
    def bisect_left(grid, val: int) -> int:
        return bisect_left(grid, _CompareFirstTuple(val, None))

    @staticmethod
    def bisect_right(grid, val: int) -> int:
        return bisect_right(grid, _CompareFirstTuple(val, None))


# These lists undergo inserts, appends, and removals fairly often. If they become a bottleneck,
#   another data structure (e.g. tree) would give better performance for those operations.
_GridYAxisT = List[_CompareFirstTuple[Set[DataId]]]
_GridXAxisT = List[_CompareFirstTuple[_GridYAxisT]]


class _DataRoot:
    """
    A root data model which may have other data models mapped into it
    """

    @property
    def length(self) -> int:
        return len(self.data)

    def __init__(self, model: DataModel, data: bytes):
        self.model: DataModel = model
        self.data = data
        self._children: Dict[DataId, DataModel] = dict()

        # A pair of sorted 2D arrays, where each "point" in the grid is a set of children's data IDs
        # The (X, Y) coordinates of each grid correspond to the range of these children
        # The coordinates are flipped between these 2 grids:
        #   (x,y) in _grid_starts_first is (y,x) in _grid_ends_first
        #   The same set object is shared by both points
        # Sometimes it is more efficient to search in one grid or the other
        self._grid_starts_first: _GridXAxisT = []  # X axis is range starts, Y axis is ends
        self._grid_ends_first: _GridXAxisT = []  # X axis is range ends, Y axis is starts

    def get_children(self) -> Iterable[DataModel]:
        return self._children.values()

    def add_mapped_model(self, model: DataModel):
        if model.range.start < 0 or model.range.end > self.length:
            raise OutOfBoundError(
                f"New mapped data model {model.id.hex()} is outside the bounds of its root "
                f"{self.model.id.hex()}: ({model.range} is outside of {self.model.range})"
            )

        self._children[model.id] = model

        # if there is not already a set at the coords of model.range, a new set is created
        # _grid_starts_first and _grid_ends_first must share this set
        # So we create it outside, then pass it into _add_to_grid, in case it is needed
        default_set: Set[DataId] = set()
        set1 = self._add_to_grid(
            self._grid_starts_first, model.range.start, model.range.end, default_set
        )
        set2 = self._add_to_grid(
            self._grid_ends_first, model.range.end, model.range.start, default_set
        )

        assert set1 is set2  # These sets should always be the same!
        set1.add(model.id)

    def delete_mapped_model(self, model: DataModel):
        if model.id not in self._children:
            raise NotFoundError(
                f"Data model with ID {model.id.hex()} is not a child of {self.model.id.hex()}"
            )

        ids_at_point = self._get_ids_at(model.range.start, model.range.end)
        ids_at_point.remove(model.id)
        if len(ids_at_point) == 0:
            self._remove_from_grid(self._grid_starts_first, model.range.start, model.range.end)
            self._remove_from_grid(self._grid_ends_first, model.range.end, model.range.start)

        del self._children[model.id]

    def resize_range(self, resized_range: Range, size_diff: int):
        try:
            for shifted_child_id, (start_shift, end_shift) in self._resize_range(
                resized_range, size_diff
            ):
                shifted_child = self._children[shifted_child_id]
                shifted_child.range = Range(
                    shifted_child.range.start + start_shift,
                    shifted_child.range.end + end_shift,
                )

        except _ShiftBreaksSortError:
            raise PatchOverlapError(
                "Cannot resize child overlapping with the boundaries of other children!"
            )

        self.model.range = Range(0, self.model.range.end + size_diff)

    def get_children_with_boundaries_intersecting_range(self, r: Range) -> List[DataModel]:
        intersecting_model_ids: Set[DataId] = set()

        for starts_in_range in self._get_ids_in_range(
            start_range=(r.start, r.end), start_inclusivity=(False, False)
        ):
            intersecting_model_ids.add(starts_in_range)

        for ends_in_range in self._get_ids_in_range(
            end_range=(r.start, r.end), end_inclusivity=(False, False)
        ):
            intersecting_model_ids.add(ends_in_range)

        return [self._children[data_id] for data_id in intersecting_model_ids]

    def get_children_affected_by_ranges(
        self, patch_ranges: List[Range]
    ) -> Iterable[Tuple[DataId, Range]]:
        children_overlapping_ranges: Dict[Range, Iterable[DataId]] = defaultdict(set)
        for patch_range in patch_ranges:
            children_overlapping_ranges[patch_range] = self._get_ids_in_range(
                start_range=(None, patch_range.end),
                end_range=(patch_range.start, None),
                start_inclusivity=(False, False),
                end_inclusivity=(False, False),
            )

        for patched_range, overlapping_data_ids in children_overlapping_ranges.items():
            for data_id in overlapping_data_ids:
                model = self._children[data_id]
                yield data_id, patched_range.intersect(model.range).translate(-model.range.start)

    def _get_ids_at(self, start: int, end: int) -> Set[DataId]:
        i = _CompareFirstTuple.bisect_left(self._grid_starts_first, start)
        if i >= len(self._grid_starts_first):
            raise KeyError()
        fetched_start: int
        column: _GridYAxisT
        fetched_start, column = self._grid_starts_first[i]
        if fetched_start != start:
            raise KeyError()

        j = _CompareFirstTuple.bisect_left(column, end)
        if j >= len(column) or column[j][0] != end:
            raise KeyError()

        return column[j][1]

    def _get_ids_in_range(
        self,
        start_range: Tuple[Optional[int], Optional[int]] = (None, None),
        end_range: Tuple[Optional[int], Optional[int]] = (None, None),
        start_inclusivity: Tuple[bool, bool] = (True, False),
        end_inclusivity: Tuple[bool, bool] = (True, False),
    ) -> Iterable[DataId]:
        start_min, start_max = start_range
        end_min, end_max = end_range

        if start_range != (None, None):
            starts_iter = self._iter_grid_axis(
                self._grid_starts_first, start_min, start_max, start_inclusivity
            )
            for _, ends in starts_iter:
                ends_iter = self._iter_grid_axis(ends, end_min, end_max, end_inclusivity)
                for _, vals in ends_iter:
                    yield from vals
        elif end_range != (None, None):
            ends_iter = self._iter_grid_axis(
                self._grid_ends_first, end_min, end_max, end_inclusivity
            )
            for _, starts in ends_iter:
                starts_iter = self._iter_grid_axis(starts, start_min, start_max, start_inclusivity)
                for _, vals in starts_iter:
                    yield from vals
        else:
            for _, column in self._grid_starts_first:
                for _, vals in column:
                    yield from vals

    def _resize_range(
        self, resized_range: Range, size_diff: int
    ) -> Iterable[Tuple[DataId, Tuple[int, int]]]:
        # Update the _grid_ends_first
        for ids_ending_after_range in self._shift_grid_axis(
            self._grid_ends_first,
            size_diff,
            merge_func=self._merge_columns,
            minimum=resized_range.end,
            inclusive=(True, False) if resized_range.length() != 0 else (False, False),
        ):
            for _, ids_starting_before_range in self._iter_grid_axis(
                ids_ending_after_range,
                maximum=resized_range.end,
            ):
                for model_id in ids_starting_before_range:
                    # Only end shifted
                    yield model_id, (0, size_diff)
            for ids_entirely_after_range in self._shift_grid_axis(
                ids_ending_after_range,
                size_diff,
                merge_func=set.union,
                minimum=resized_range.end,
            ):
                for model_id in ids_entirely_after_range:
                    # Both start and end shifted
                    yield model_id, (size_diff, size_diff)

        # Update the _grid_starts_first
        for _, ids_starting_before_range in self._iter_grid_axis(
            self._grid_starts_first,
            maximum=resized_range.start,
            inclusive=(True, True) if resized_range.length() != 0 else (True, False),
        ):
            for _ in self._shift_grid_axis(
                ids_starting_before_range,
                size_diff,
                merge_func=set.union,
                minimum=resized_range.start,
                inclusive=(False, False),
            ):
                # Only end shifted
                pass

        for ends in self._shift_grid_axis(
            self._grid_starts_first,
            size_diff,
            merge_func=self._merge_columns,
            minimum=resized_range.end,
            inclusive=(True, False),
        ):
            for _ in self._shift_grid_axis(
                ends,
                size_diff,
                merge_func=set.union,
            ):
                # Both start and end shifted
                pass

    @staticmethod
    def _add_to_grid(
        grid: _GridXAxisT,
        x: int,
        y: int,
        default: Set[DataId],
    ) -> Set[DataId]:
        i = _CompareFirstTuple.bisect_left(grid, x)
        if i >= len(grid) or grid[i][0] != x:
            column: _GridYAxisT = []
            grid.insert(i, _CompareFirstTuple(x, column))
        else:
            column = grid[i][1]

        j = _CompareFirstTuple.bisect_left(column, y)
        if j >= len(column) or column[j][0] != y:
            column.insert(j, _CompareFirstTuple(y, default))

        return column[j][1]

    @staticmethod
    def _remove_from_grid(grid: _GridXAxisT, x: int, y: int):
        i = _CompareFirstTuple.bisect_left(grid, x)
        if i >= len(grid) or grid[i][0] != x:
            raise KeyError(f"No column {x} in the grid!")
        else:
            column = grid[i][1]

        j = _CompareFirstTuple.bisect_left(column, y)
        if j >= len(column) or column[j][0] != y:
            raise KeyError(f"No point {y} in column {x} in the grid!")

        column.pop(j)
        if len(column) == 0:
            grid.pop(i)

    @staticmethod
    def _iter_grid_axis(
        grid: List[T],
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        inclusive: Tuple[bool, bool] = (True, False),
    ) -> Iterable[T]:
        if minimum is not None:
            if inclusive[0]:
                min_i = _CompareFirstTuple.bisect_left(grid, minimum)
            else:
                min_i = _CompareFirstTuple.bisect_right(grid, minimum)
        else:
            min_i = 0

        if maximum is not None:
            if inclusive[1]:
                max_i = _CompareFirstTuple.bisect_right(grid, maximum)
            else:
                max_i = _CompareFirstTuple.bisect_left(grid, maximum)
        else:
            max_i = len(grid)

        i = min_i
        while i < max_i:
            yield grid[i]
            i += 1

    @staticmethod
    def _shift_grid_axis(
        axis: List[_CompareFirstTuple[T]],
        shift: int,
        merge_func: Callable[[T, T], T],
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        inclusive: Tuple[bool, bool] = (True, False),
    ) -> Iterable[T]:
        """
        Shift a range of values in an axis, without affecting the sorted order of the points in
        the axis. With two exceptions:
        - If the minimum shifted point is shifted DOWN exactly enough to be equal to the previous
          point (which has by definition not been shifted), those two points are allowed to merge
        - If the maximum shifted point is shifted UP exactly enough to be equal to the next
          point (which has by definition not been shifted), those two points are allowed to merge

        At most one of these can happen when shifting. The `merge_func` parameter handles merging
        those two points. Since we may be shifting either a row or a column, the merged "points" may
        be either columns (if shifting rows) or sets of bytes (if shifting columns).
        """
        pre_yield = None
        post_yield = None

        if minimum is not None:
            if inclusive[0]:
                min_i = _CompareFirstTuple.bisect_left(axis, minimum)
            else:
                min_i = _CompareFirstTuple.bisect_right(axis, minimum)
        else:
            min_i = 0

        if 0 < min_i < (len(axis) - 1):
            post_shift_min = axis[min_i][0] + shift
            if post_shift_min < axis[min_i - 1][0]:
                raise _ShiftBreaksSortError(
                    f"shifting {minimum} to {maximum} by {shift} would collide at the lower range!"
                )
            elif post_shift_min == axis[min_i - 1][0]:
                # will merge the lowest val in shifted range into previous
                val1 = axis[min_i - 1][1]
                _, pre_yield = axis.pop(min_i)

        if maximum is not None:
            if inclusive[1]:
                max_i = _CompareFirstTuple.bisect_left(axis, maximum)
            else:
                max_i = _CompareFirstTuple.bisect_right(axis, maximum)
        else:
            max_i = len(axis)

        if 0 < (max_i + 1) < len(axis):
            post_shift_max = axis[max_i][0] + shift
            if post_shift_max > axis[max_i + 1][0]:
                raise _ShiftBreaksSortError(
                    f"shifting {minimum} to {maximum} by {shift} would collide at the upper range!"
                )
            elif post_shift_max == axis[max_i + 1][0]:
                # will merge the highest val in shifted range into next
                val1 = axis[max_i + 1][1]
                _, post_yield = axis.pop(max_i)

                max_i -= 1

        if pre_yield is not None:
            yield pre_yield
            axis[min_i - 1] = _CompareFirstTuple(post_shift_min, merge_func(val1, pre_yield))

        i = min_i
        while i < max_i:
            old_key, val = axis[i]
            axis[i] = _CompareFirstTuple(old_key + shift, val)
            yield val
            i += 1

        if post_yield is not None:
            yield post_yield
            axis[max_i + 2] = _CompareFirstTuple(post_shift_max, merge_func(val1, post_yield))

    @staticmethod
    def _merge_columns(
        col1: _GridYAxisT,
        col2: _GridYAxisT,
    ) -> _GridYAxisT:
        merged_columns: _GridYAxisT = []
        for key, _vals in itertools.groupby(heapq.merge(col1, col2), key=lambda x: x[0]):
            vals = tuple(v for _, v in _vals)
            if len(vals) == 1:
                val: Set[DataId] = vals[0]
            else:
                val = set.union(*vals)

            merged_columns.append(_CompareFirstTuple(key, val))

        return merged_columns


class _PatchResizeTracker:
    def __init__(self):
        # Map ORIGINAL offsets in
        self.resizing_shifts: SortedList[List[int]] = SortedList(key=lambda x: x[0])
        self.resized_ranges = []
        self.resizing_shifts.add([0, 0])

    def overlaps_resized_range(self, r: Range) -> bool:
        return any(
            resized_range.overlaps(r) or resized_range.start == r.start
            for resized_range in self.resized_ranges
        )

    def translate_range(self, r: Range) -> Range:
        return Range(
            self.get_shifted_point(r.start),
            self.get_shifted_point(r.end),
        )

    def get_shifted_point(self, point: int) -> int:
        i = self.resizing_shifts.bisect_right((point, 0))
        assert i != 0
        previous_shift_end, shift = self.resizing_shifts[i - 1]
        return point + shift

    def add_new_resized_range(self, r: Range, size_diff: int):
        self.resized_ranges.append(r)
        i = self.resizing_shifts.bisect_right((r.end, 0))
        total_offset_here = self.resizing_shifts[i - 1][1] + size_diff
        for node in self.resizing_shifts.islice(i):
            node[1] += size_diff
        self.resizing_shifts.add([r.end, total_offset_here])

    def get_total_size_diff(self) -> int:
        return self.resizing_shifts[-1][1]



from abc import ABC, abstractmethod

from ofrak.service.abstract_ofrak_service import AbstractOfrakService


class IDServiceInterface(AbstractOfrakService, ABC):
    """
    The `IDServiceInterface` is an interface for a service that generates unique IDs.

    This interface is intended to be a singleton used to create IDs locally; if more than one
    instance of this service is going to be instantiated, then this service must be stateless and
    not rely on any internal state.
    """

    @abstractmethod
    def generate_id(self) -> bytes:
        """
        Generate a unique ID.

        This method guarantees that no two IDs it returns will be identical for the lifetime of
        the instantiated `IDServiceInterface`.

        :return: The unique ID that was generated.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def generate_id_from_base(base_id: bytes, key: str) -> bytes:
        """
        Generate an ID based on a base ID and key.

        This method returns the same ID each time it is called with the exact same parameters. It
        should therefore be used with caution: the callee needs to know that the parameters
        passed to this method are unique.

        :param bytes base_id: An ID used to derive the generated ID. This ID can be assumed to have
            been generated by the `generate_id` method.
        :param str key: A key used to guarantee that the generated ID is unique. The callee is
            responsible for ensuring that this key is unique.
        :return: The ID that was generated.
        """
        raise NotImplementedError()



from argparse import Namespace
from typing import Iterable, Set

from ofrak.cli.ofrak_cli import OfrakCommand, OFRAKEnvironment


class ListCommand(OfrakCommand):
    def create_parser(self, ofrak_subparsers):
        list_parser = ofrak_subparsers.add_parser(
            "list",
            help="List installed OFRAK modules and/or components.",
            description="List installed OFRAK modules and/or components. By default, prints all "
            "installed components, organized by module (equivalent to `--packages --components` "
            "flags)",
        )
        list_parser.add_argument(
            "--packages",
            "-p",
            action="store_true",
            help="List installed OFRAK packages",
        )
        list_parser.add_argument(
            "--components",
            "-c",
            action="store_true",
            help="List installed OFRAK components",
        )
        return list_parser

    def run(self, ofrak_env: OFRAKEnvironment, args: Namespace):
        output_lines = []

        if args.components or args.packages:
            list_components = args.components
            list_packages = args.packages
        else:
            list_components = True
            list_packages = True

        indent = 0
        for mod in ofrak_env.packages.values():
            if list_packages:
                output_lines.append("\t" * indent + mod.__name__)
                indent += 1
            for component in ofrak_env.components_by_package[mod]:
                if list_components:
                    output_lines.append("\t" * indent + component.__name__)
            if list_packages:
                indent -= 1

        _print_lines_without_duplicates(output_lines)


def _print_lines_without_duplicates(output_lines: Iterable[str]):
    # strip duplicates, resetting the memory of duplicates when indentation changes
    prev_indent = 0
    seen: Set[str] = set()
    for line in output_lines:
        indent = line.rfind("\t") + 1
        if indent != prev_indent:
            seen = set()
        prev_indent = indent
        if line not in seen:
            print(line)
        seen.add(line)



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Iterable, Any, Tuple, Optional

from ofrak.model.resource_model import (
    ResourceModel,
    ResourceIndexedAttribute,
    ResourceModelDiff,
)
from ofrak.model.tag_model import ResourceTag
from ofrak.service.abstract_ofrak_service import AbstractOfrakService


class ResourceServiceWalkError(RuntimeError):
    pass


class ResourceFilterCondition(Enum):
    AND = 1
    OR = 2


@dataclass
class ResourceAttributeFilter:
    attribute: ResourceIndexedAttribute


@dataclass
class ResourceAttributeRangeFilter(ResourceAttributeFilter):
    """
    A resource's [index][ofrak.model.resource_model.index] value must be within a range. The range
    must be bounded with either a maximum or minimum. The maximum is exclusive, i.e. if the index
    value is equal to the max, then the resource is excluded.
    """

    min: Any = None
    max: Any = None


@dataclass
class ResourceAttributeValueFilter(ResourceAttributeFilter):
    value: Any


@dataclass
class ResourceAttributeValuesFilter(ResourceAttributeFilter):
    values: Tuple[Any, ...]


@dataclass
class ResourceFilter:
    include_self: bool = False
    tags: Optional[Iterable[ResourceTag]] = None
    tags_condition: ResourceFilterCondition = ResourceFilterCondition.AND
    attribute_filters: Optional[Iterable[ResourceAttributeFilter]] = None

    @classmethod
    def with_tags(cls, *tags: ResourceTag):
        return ResourceFilter(tags=tags)


class ResourceSortDirection(Enum):
    DESCENDANT = 0
    ASCENDANT = 1


@dataclass
class ResourceSort:
    attribute: ResourceIndexedAttribute
    direction: ResourceSortDirection = ResourceSortDirection.ASCENDANT


class ResourceServiceInterface(AbstractOfrakService, metaclass=ABCMeta):
    """
    Stores [ResourceModels][ofrak.model.resource_model.ResourceModel] in a tree structure and
    provides methods to walk that tree according to given
    [ResourceFilters][ofrak.service.resource_service_i.ResourceFilter]. Resource instantiation is
    handled by the [ResourceFactory][ofrak.resource.ResourceFactory].
    """

    @abstractmethod
    async def create(self, resource: ResourceModel) -> ResourceModel:
        """
        Add a [ResourceModel][ofrak.model.resource_model.ResourceModel] to the resource service
        database according to the given model. If the ``resource`` model says it has a parent,
        ``resource`` will be added as a child of that parent.

        :param resource: The resource model to add to the database

        :raises AlreadyExistError: If ``resource`` has an ID which already exists in the database
        :raises NotFoundError: If ``resource`` has a parent ID but no resource with that ID exists

        :return: The same model which was passed in, with no changes

        """
        raise NotImplementedError()

    @abstractmethod
    async def get_root_resources(self) -> Iterable[ResourceModel]:
        """
        Get all of the root resources known to this resource service. Any resource created
        without a parent will be returned by this method.

        :return: All resources with no parents
        """
        raise NotImplementedError()

    @abstractmethod
    async def verify_ids_exist(self, resource_ids: Iterable[bytes]) -> Iterable[bool]:
        """
        Check if a number of resource IDs exist in the resource store. This is useful for
        filtering out IDs of resources which have been deleted.

        :param resource_ids: Iterable of resource IDs to check for

        :return: A boolean for each resource ID, True if it exists in the store and False otherwise
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_by_data_ids(self, data_ids: Iterable[bytes]) -> Iterable[ResourceModel]:
        """
        Get the resource models with a given sequence of data IDs.

        :param data_ids: A list of valid data IDs

        :raises NotFoundError: If there is not a resource for all of the IDs in `data_ids`

        :return: A sequence of resource models each with one of the given data IDs, in the same
        order which `data_ids` was given in.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_by_ids(self, resource_ids: Iterable[bytes]) -> Iterable[ResourceModel]:
        """
        Get the resource models with a given sequence of resource IDs.

        :param resource_ids: A list of valid resource IDs

        :raises NotFoundError: If there is not a resource for all of the IDs in `resource_ids`

        :return: A sequence of resource models each with one of the given resource IDs, in the same
        order which `resource_ids` was given in.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_by_id(self, resource_id: bytes) -> ResourceModel:
        """
        Get the resource model with a given resource ID.

        :param resource_id: A valid resource ID

        :raises NotFoundError: If there is not a resource with resource ID `resource_id`

        :return: The resource model with ID matching `resource_id`
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_depths(self, resource_ids: Iterable[bytes]) -> Iterable[int]:
        """
        Get the depth of each resource in `resource_ids`.

        :param resource_ids: A list of valid resource IDs

        :raises NotFoundError: If there is not a resource for all of the IDs in `resource_ids`

        :return: A sequence of resource model depths, in the same order which `resource_ids`
        was given in.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_ancestors_by_id(
        self,
        resource_id: bytes,
        max_count: int = -1,
        r_filter: Optional[ResourceFilter] = None,
    ) -> Iterable[ResourceModel]:
        """
        Get the resource models of the ancestors of a resource with a given ID. These ancestors
        may be filtered by an optional filter argument. A maximum count of ancestors may also be
        given, to cap the number of (filtered or unfiltered) ancestors returned.

        :param resource_id: ID of resource to get ancestors of
        :param max_count: Optional argument to cap the number of models returned; if set to -1
        (default) then any number of ancestors may be returned
        :param r_filter: Optional resource filter for the resource models returned; if set to
        `None`, all ancestors may be returned (the model for `resource_id` is excluded),
        otherwise all ancestors matching the filter may be returned (possibly including the model
        for `resource_id`), up to the maximum allowed by `max_count`

        :raises NotFoundError: If there is not a resource with resource ID `resource_id`

        :return: As many ancestors of `resource_id` matching `r_filter` as `max_count`
        allows, in order of reverse depth (deeper resources first, root last)
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_descendants_by_id(
        self,
        resource_id: bytes,
        max_count: int = -1,
        max_depth: int = -1,
        r_filter: Optional[ResourceFilter] = None,
        r_sort: Optional[ResourceSort] = None,
    ) -> Iterable[ResourceModel]:
        """
        Get the resource models of the descendants of a resource with a given ID. These descendants
        may be filtered by an optional filter argument. A maximum count of descendants may also be
        given, to cap the number of (filtered or unfiltered) descendants returned. A maximum
        depth may also be given, to limit how deep to search for descendants.

        :param resource_id: ID of resource to get descendants of
        :param max_count: Optional argument to cap the number of models returned; if set to -1
        (default) then any number of descendants may be returned
        :param max_depth: Optional argument to limit the depth to search for descendants; if set
        to -1 (default) then descendants of any depth may be returned; if set to 1, for example,
        only children of `resource_id` may be returned
        :param r_filter: Optional resource filter for the resource models returned; if set to
        `None` all descendants may be returned (the model for `resource_id` is excluded),
        otherwise all descendants matching the filter may be returned (possibly including the model
        for `resource_id`), up to the maximum allowed by `max_count`
        :param r_sort: Optional logic to order the returned descendants by the value of a
        specific attribute of each descendant

        :raises NotFoundError: If there is not a resource with resource ID `resource_id`

        :return: As many descendants of `resource_id` matching `r_filter` as `max_count`
        allows, in order specified by `r_sort`. If `r_sort` is None, no specific ordering is
        guaranteed.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_siblings_by_id(
        self,
        resource_id: bytes,
        max_count: int = -1,
        r_filter: Optional[ResourceFilter] = None,
        r_sort: Optional[ResourceSort] = None,
    ) -> Iterable[ResourceModel]:
        """
        Get the resource models of the siblings of a resource with a given ID. These siblings
        may be filtered by an optional filter argument. A maximum count of siblings may also be
        given, to cap the number of (filtered or unfiltered) siblings returned.

        :param resource_id: ID of resource to get siblings of
        :param max_count: Optional argument to cap the number of models returned; if set to -1
        (default) then any number of siblings may be returned
        :param r_filter: Optional resource filter for the resource models returned; if set to
        None all siblings may be returned (the model for `resource_id` is excluded),
        otherwise all siblings matching the filter may be returned (possibly including the model
        for `resource_id`), up to the maximum allowed by `max_count`
        :param r_sort: Optional logic to order the returned siblings by the value of a
        specific attribute of each sibling

        :raises NotFoundError: If there is not a resource with resource ID `resource_id`
        :raises NotFoundError: If the resource with ID `resource_id` does not have siblings
        because it is a root


        :return: As many siblings of `resource_id` matching `r_filter` as `max_count`
        allows, in order specified by `r_sort`. If `r_sort` is None, no specific ordering is
        guaranteed.
        """
        raise NotImplementedError()

    @abstractmethod
    async def update(self, resource_diff: ResourceModelDiff) -> ResourceModel:
        """
        Modify a stored resource model according to the differences in the given diff object.

        :param resource_diff: Diff object containing changes to a resource model, as well as the
        resource ID of the model to update

        :raises NotFoundError: If there is not a resource with resource ID matching the ID in
        `resource_diff`

        :return: The updated resource model (with changes applied)
        """
        raise NotImplementedError()

    @abstractmethod
    async def update_many(
        self, resource_diffs: Iterable[ResourceModelDiff]
    ) -> Iterable[ResourceModel]:
        """
        Modify a stored resource model according to the differences in the given diff object.

        :param resource_diffs: Diff objects containing changes to resource models, as well as the
        resource ID of each model to update

        :raises NotFoundError: If there is not a resource with resource ID matching one of the IDs
        in `resource_diffs`

        :return: The updated resource models (with changes applied)
        """
        raise NotImplementedError()

    @abstractmethod
    async def rebase_resource(self, resource_id: bytes, new_parent_id: bytes):
        """
        Move a resource which was a child to instead be a child of a different resource.

        :param resource_id: resource ID of the resource to rebase
        :param new_parent_id: resource ID of the new parent resource for `resource_id`

        :raises NotFoundError: If there is not a resource with resource ID `resource_id`
        :raises NotFoundError: If there is not a resource with resource ID `new_parent_id`
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_resource(self, resource_id: bytes) -> Iterable[ResourceModel]:
        """
        Delete a resource by ID and all of its descendants, removing them from the database. If no
        resource for the given ID is found, it is assumed the resource has already been deleted
        (does not raise an error).

        :param resource_id: The ID of the resource to delete

        :return: all of the models that were deleted
        """
        raise NotImplementedError()

    @abstractmethod
    async def delete_resources(self, resource_ids: Iterable[bytes]) -> Iterable[ResourceModel]:
        """
        Delete multiple resources by ID and all of their descendants, removing them from the
        database. If no resource for any given ID is found, it is assumed the resource has already
        been deleted (does not raise an error).

        :param resource_ids: The ID of the resources to delete

        :return: all of the models that were deleted
        """
        raise NotImplementedError()



from abc import ABC
from typing import Tuple

from ofrak import Analyzer, Unpacker, Resource


class MockAnalyzer(Analyzer[None, Tuple], ABC):
    def __init__(self):
        super().__init__(None, None, None)

    async def analyze(self, resource: Resource, config=None) -> Tuple:
        return ()


class MockUnpacker(Unpacker[None]):
    targets = ()
    children = ()

    def __init__(self):
        super().__init__(None, None, None, None)

    async def unpack(self, resource, config=None):
        pass



import asyncio
import tempfile
from dataclasses import dataclass
import logging
from subprocess import CalledProcessError

from ofrak import Identifier, Analyzer
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.core import PY_LZO_TOOL
from ofrak.resource import Resource
from ofrak.core.filesystem import File, Folder, FilesystemRoot, SpecialFileType
from ofrak.core.binary import GenericBinary

from ofrak.model.component_model import ComponentExternalTool
from ofrak_type.range import Range

try:
    from ubireader.ubifs.defines import UBIFS_NODE_MAGIC

    from ubireader import ubi_io
    from ubireader.ubifs import ubifs as ubireader_ubifs
    from ubireader.ubifs.defines import PRINT_UBIFS_KEY_HASH, PRINT_UBIFS_COMPR
    from ubireader.utils import guess_leb_size
except ModuleNotFoundError:
    pass

LOGGER = logging.getLogger(__name__)

MKFS_UBIFS_TOOL = ComponentExternalTool(
    "mkfs.ubifs",
    "http://www.linux-mtd.infradead.org/faq/ubifs.html",
    install_check_arg="--help",
    apt_package="mtd-utils",
    brew_package=None,  # This isn't compatible with macos, but there may be an alternative tool to do the bidding.
)


@dataclass
class SuperblockNode:
    """
    Each UBIFS image has a superblock which describe a large number of parameters regarding the filesystem. The
    minimal set of parameters necessary to re-pack an UBIFS filesystem are stored here.
    (see also: https://elixir.bootlin.com/linux/v6.1.7/source/fs/ubifs/ubifs.h#L1017).


    :var max_leb_count: Maximum number / limit of Logical Erase Blocks
    :var default_compr: Default compression algorithm
    :var fanout: Fanout of the index tree (number of links per indexing node)
    :var key_hash: Type of hash function used for keying direntries (typically 'r5' of reiserfs)
    :var orph_lebs: Number of LEBs used for orphan area (orphans are inodes with no links; see https://elixir.bootlin.com/linux/v6.1.7/source/fs/ubifs/orphan.c#L13)
    :var log_lebs: LEBs reserved for the journal (see the 'Journal' section in https://www.kernel.org/doc/Documentation/filesystems/ubifs-authentication.rst)
    """

    max_leb_count: int
    default_compr: str  # PRINT_UBIFS_COMPR
    fanout: int
    key_hash: str  # PRINT_UBIFS_KEY_HASH
    orph_lebs: int
    log_lebs: int


@dataclass
class Ubifs(GenericBinary, FilesystemRoot):
    """
    UBIFS is a filesystem specially made to run on top of an UBI volume layer. UBIFS specifically provides
    indexing, compression, encryption / authentication and some other filesystem-related features.

    As part of an UBI image, re-packing an UBIFS image requires the 'min_io_size' and 'leb_size' properties that
    are stored as part of the UBI header (https://elixir.bootlin.com/linux/v6.1.7/source/drivers/mtd/ubi/ubi.h#L441).

    Each UBIFS image has a superblock which is encoded in OFRAK as a SuperblockNode.

    Some documentation about UBIFS layout can also be found here:
    https://www.kernel.org/doc/Documentation/filesystems/ubifs-authentication.rst
    http://www.linux-mtd.infradead.org/doc/ubifs.html

    :var min_io_size: Minimum number of bytes per I/O transaction (see http://www.linux-mtd.infradead.org/doc/ubi.html#L_min_io_unit)
    :var leb_size: Size of Logical Erase Blocks
    :var superblock: A SuberblockNode

    """

    min_io_size: int
    leb_size: int
    superblock: SuperblockNode


class UbifsAnalyzer(Analyzer[None, Ubifs]):
    """
    Extract UBIFS parameters required for packing a resource.
    """

    targets = (Ubifs,)
    outputs = (Ubifs,)

    external_dependencies = (PY_LZO_TOOL,)

    async def analyze(self, resource: Resource, config=None) -> Ubifs:
        with tempfile.NamedTemporaryFile() as temp_file:
            resource_data = await resource.get_data()
            temp_file.write(resource_data)
            temp_file.flush()

            ubifs_obj = ubireader_ubifs(
                ubi_io.ubi_file(
                    temp_file.name,
                    block_size=guess_leb_size(temp_file.name),
                    start_offset=0,
                    end_offset=None,
                )
            )
            return Ubifs(
                ubifs_obj._get_min_io_size(),
                ubifs_obj._get_leb_size(),
                SuperblockNode(
                    ubifs_obj.superblock_node.max_leb_cnt,
                    PRINT_UBIFS_COMPR[ubifs_obj.superblock_node.default_compr],
                    ubifs_obj.superblock_node.fanout,
                    PRINT_UBIFS_KEY_HASH[ubifs_obj.superblock_node.key_hash],
                    ubifs_obj.superblock_node.orph_lebs,
                    ubifs_obj.superblock_node.log_lebs,
                ),
            )


class UbifsUnpacker(Unpacker[None]):
    """
    Unpack the UBIFS image into a filesystem representation.
    """

    targets = (Ubifs,)
    children = (File, Folder, SpecialFileType)

    external_dependencies = (PY_LZO_TOOL,)

    async def unpack(self, resource: Resource, config=None):
        with tempfile.TemporaryDirectory() as temp_flush_dir:
            # flush to disk
            with open(f"{temp_flush_dir}/input.img", "wb") as temp_file:
                resource_data = await resource.get_data()
                temp_file.write(resource_data)
                temp_file.flush()

            cmd = [
                "ubireader_extract_files",
                "-k",
                "-o",
                f"{temp_flush_dir}/output",
                temp_file.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)

            ubifs_view = await resource.view_as(Ubifs)
            await ubifs_view.initialize_from_disk(f"{temp_flush_dir}/output")


class UbifsPacker(Packer[None]):
    """
    Generate an UBIFS image from a filesystem representation in OFRAK.
    """

    targets = (Ubifs,)
    external_dependencies = (MKFS_UBIFS_TOOL,)

    async def pack(self, resource: Resource, config=None) -> None:
        ubifs_view = await resource.view_as(Ubifs)
        flush_dir = await ubifs_view.flush_to_disk()

        with tempfile.NamedTemporaryFile(mode="rb") as temp:
            cmd = [
                "mkfs.ubifs",
                "-m",
                f"{ubifs_view.min_io_size}",
                "-e",
                f"{ubifs_view.leb_size}",
                "-c",
                f"{ubifs_view.superblock.max_leb_count}",
                "-x",
                f"{ubifs_view.superblock.default_compr}",
                "-f",
                f"{ubifs_view.superblock.fanout}",
                "-k",
                f"{ubifs_view.superblock.key_hash}",
                "-p",
                f"{ubifs_view.superblock.orph_lebs}",
                "-l",
                f"{ubifs_view.superblock.log_lebs}",
                "-F",
                "-r",
                flush_dir,
                temp.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)
            new_data = temp.read()

            resource.queue_patch(Range(0, await resource.get_data_length()), new_data)


class UbifsIdentifier(Identifier):
    """
    Check the first four bytes of a resource and tag the resource as Ubifs if it matches the file magic.
    """

    targets = (File, GenericBinary)

    external_dependencies = (PY_LZO_TOOL,)

    async def identify(self, resource: Resource, config=None) -> None:
        datalength = await resource.get_data_length()
        if datalength >= 4:
            data = await resource.get_data(Range(0, 4))
            if data == UBIFS_NODE_MAGIC:
                resource.add_tag(Ubifs)



import os.path
from typing import Optional

from ofrak import OFRAKContext, Resource

import argparse
from pathlib import Path
import time
import sys

from ofrak.cli.ofrak_cli import OfrakCommandRunsScript
from ofrak.core import FilesystemEntry
from ofrak.gui.server import open_gui


class UnpackCommand(OfrakCommandRunsScript):
    def __init__(self):
        self._filename_trackers = dict()
        self._resource_paths = dict()

    def create_parser(self, parser: argparse._SubParsersAction):
        subparser = parser.add_parser(
            "unpack",
            help="Unpack all identified structures that can be unpacked with OFRAK",
            description="Import a file as an OFRAK resource, then identifies and unpacks it. The "
            "resource's children are written to the output directory as individual "
            "files. Children which have no data are not written as files. A file `__ofrak_info__` "
            "is also written to the output directory, containing the known OFRAK tags and "
            "attributes for each descendant.",
        )
        subparser.add_argument(
            "-o",
            "--output_directory",
            help="Directory to write unpacked resource tree to. If no directory is given, a new one"
            " will be created in the same directory as the file being unpacked.",
        )
        subparser.add_argument(
            "--recursive",
            "-r",
            action="store_true",
            default=False,
            help="Unpack recursively: all resources unpacked from the root will be unpack, as "
            "well as all resources unpacked from those, and so.",
        )
        subparser.add_argument("filename", help="File to unpack")

        # GUI args
        subparser.add_argument(
            "--gui",
            action="store_true",
            help="Open the OFRAK GUI after unpacking",
            default=False,
        )
        subparser.add_argument(
            "-gH",
            "--gui-hostname",
            action="store",
            help="Set GUI server host address.",
            default="127.0.0.1",
        )
        subparser.add_argument(
            "-gp",
            "--gui-port",
            action="store",
            type=int,
            help="Set GUI server host port.",
            default=8080,
        )
        subparser.add_argument(
            "--gui-no-browser",
            action="store_true",
            help="Don't open the browser to the OFRAK GUI",
        )

        self.add_ofrak_arguments(subparser)

        return subparser

    async def ofrak_func(self, ofrak_context: OFRAKContext, args: argparse.Namespace):
        print(f"Unpacking file: {args.filename}\n")
        root_resource = await ofrak_context.create_root_resource_from_file(args.filename)

        if args.recursive:
            await root_resource.unpack_recursively()
        else:
            await root_resource.unpack()

        if args.output_directory:
            extraction_dir = Path(args.output_directory)
        else:
            file_path = Path(args.filename)
            parent_dir = file_path.parent
            extraction_dir = Path(
                parent_dir / f'{file_path.name}_extracted_{time.strftime("%Y%m%d%H%M%S")}'
            )

        if extraction_dir.exists():
            if any(extraction_dir.iterdir()):
                print(f"Found files in {extraction_dir}, ABORTING!", file=sys.stderr)
                return
        else:
            extraction_dir.mkdir()

        print(f"Extracting data to {extraction_dir}")

        root_resource_path = os.path.join(
            extraction_dir,
            await self.get_filesystem_name(root_resource),
        )
        info_dump_path = os.path.join(extraction_dir, "__ofrak_info__")
        await self.resource_tree_to_files(root_resource, root_resource_path)

        info_dump = await root_resource.summarize_tree(
            summarize_resource_callback=lambda resource: _custom_summarize_resource(resource, self)
        )
        # Some characters in filename bytestrings are no valid unicode, can't be printed, must be replaced
        # https://stackoverflow.com/questions/27366479/python-3-os-walk-file-paths-unicodeencodeerror-utf-8-codec-cant-encode-s
        info_dump = info_dump.encode("utf-8", "replace").decode(
            "utf-8",
        )

        with open(info_dump_path, "w", encoding="utf-8") as f:
            f.write(info_dump)

        print(info_dump)

        if args.gui:
            server = await open_gui(
                args.gui_hostname,
                args.gui_port,
                focus_resource=root_resource,
                open_in_browser=(not args.gui_no_browser),
            )
            await server.run_until_cancelled()

    async def resource_tree_to_files(self, resource: Resource, path):
        children_dir = path + ".ofrak_children"
        for child_resource in await resource.get_children():
            filename = await self.get_filesystem_name(child_resource)

            if not os.path.exists(children_dir):
                os.mkdir(children_dir)

            child_path = os.path.join(children_dir, filename)
            await self.resource_tree_to_files(child_resource, child_path)

        if resource.get_data_id() is None:
            return
        data = await resource.get_data()
        if len(data) == 0:
            return
        with open(path, "wb") as f:
            f.write(data)
        self._resource_paths[resource.get_id()] = path

    async def get_filesystem_name(self, resource: Resource) -> str:
        if resource.has_tag(FilesystemEntry):
            file_view = await resource.view_as(FilesystemEntry)
            filename = file_view.name
        else:
            filename = resource.get_caption()

        parent_id = resource.get_model().parent_id
        filesystem_name_key = (parent_id, filename)
        if filesystem_name_key in self._filename_trackers:
            name_suffixes = self._filename_trackers[filesystem_name_key]
        else:
            name_suffixes = {resource.get_id(): ""}
            self._filename_trackers[filesystem_name_key] = name_suffixes

            return filename

        if resource.get_id() in name_suffixes:
            return filename + name_suffixes[resource.get_id()]
        else:
            suffix = f"_{len(name_suffixes)}"
            name_suffixes[resource.get_id()] = suffix
            return filename + suffix

    def get_path(self, resource: Resource) -> Optional[str]:
        return self._resource_paths.get(resource.get_id())


async def _custom_summarize_resource(resource: Resource, unpack_cmd: UnpackCommand) -> str:
    attributes_info = ", ".join(attrs_type.__name__ for attrs_type in resource._resource.attributes)
    name = await unpack_cmd.get_filesystem_name(resource)
    if " " in name:
        name = f"'{name}'"

    if resource._resource.data_id:
        data_info = f", size={await resource.get_data_length()} bytes"
    else:
        data_info = ", no data"

    path = unpack_cmd.get_path(resource)
    if path is None:
        path_info = ", (not written)"
    else:
        path_info = f", extracted-path={unpack_cmd.get_path(resource)}"
    return f"{name}: [attributes=({attributes_info}){data_info}{path_info}]"



from dataclasses import dataclass
from typing import List, Tuple

from ofrak_type import LinkableSymbolType, MemoryPermissions, InstructionSetMode
from ofrak.core.label import LabeledAddress
from ofrak_patch_maker.toolchain.model import Segment
from ofrak_type.memory_permissions import MemoryPermissions


@dataclass
class LinkableSymbolStubInfo:
    """
    Container holding the information needed to create a stub for a LinkableSymbol.

    :var asm_prefixes: The lines to prefix an assembly stub for this symbol, usually describing
    what type of symbol this is and optionally the mode
    :var segments: Segments to extract from the stub object file.
    """

    asm_prefixes: List[str]
    segments: Tuple[Segment, ...]


@dataclass
class LinkableSymbol(LabeledAddress):
    """
    A 'symbol' in the binary that may be used for the purpose of linking when injecting a patch.
    It may be extracted from symbol information that was already in the binary, added manually,
    or inferred by some analysis.
    """

    symbol_type: LinkableSymbolType
    mode: InstructionSetMode = InstructionSetMode.NONE

    def get_stub_info(self) -> LinkableSymbolStubInfo:
        """
        Get the information about this LinkableSymbol needed to generate a stub
        """
        if self.symbol_type is LinkableSymbolType.FUNC:
            return _make_rx_stub_info(self.name, self.virtual_address, self.mode)

        elif self.symbol_type is LinkableSymbolType.RO_DATA:
            return _make_r_stub_info(self.name, self.virtual_address)

        elif self.symbol_type is LinkableSymbolType.RW_DATA:
            return _make_rw_stub_info(self.name, self.virtual_address)

        else:
            raise NotImplementedError(f"No stub info factory for {self.symbol_type.name}")


def _make_rx_stub_info(name: str, vaddr: int, mode: InstructionSetMode) -> LinkableSymbolStubInfo:
    asm_prefixes = [".section .text", f".type {name}, %function"]
    if mode is InstructionSetMode.THUMB:
        asm_prefixes.append(".thumb")
    segments = (
        # Executable stub goes in .text segment
        Segment(
            segment_name=".text",
            vm_address=vaddr,
            offset=0,
            is_entry=False,
            length=0,
            access_perms=MemoryPermissions.RX,
        ),
        # Null segments required for unused .data
        Segment(
            segment_name=".data",
            vm_address=0,
            offset=0,
            is_entry=False,
            length=0,
            access_perms=MemoryPermissions.RW,
        ),
    )
    return LinkableSymbolStubInfo(asm_prefixes, segments)


def _make_r_stub_info(name: str, vaddr: int) -> LinkableSymbolStubInfo:
    asm_prefixes = [".section .rodata", f".type {name}, %object"]
    segments = (
        # Read-only symbol goes in .rodata segment
        Segment(
            segment_name=".rodata",
            vm_address=vaddr,
            offset=0,
            is_entry=False,
            length=0,
            access_perms=MemoryPermissions.R,
        ),
        # Null segments required for unused .text and .data
        Segment(
            segment_name=".text",
            vm_address=0,
            offset=0,
            is_entry=False,
            length=0,
            access_perms=MemoryPermissions.RX,
        ),
        Segment(
            segment_name=".data",
            vm_address=0,
            offset=0,
            is_entry=False,
            length=0,
            access_perms=MemoryPermissions.RW,
        ),
    )
    return LinkableSymbolStubInfo(asm_prefixes, segments)


def _make_rw_stub_info(name: str, vaddr: int) -> LinkableSymbolStubInfo:
    asm_prefixes = [".section .data", f".type {name}, %object"]
    segments = (
        # Read-write symbol goes in .data segment
        Segment(
            segment_name=".data",
            vm_address=vaddr,
            offset=0,
            is_entry=False,
            length=0,
            access_perms=MemoryPermissions.RW,
        ),
        # Null segment required for unused .text
        Segment(
            segment_name=".text",
            vm_address=0,
            offset=0,
            is_entry=False,
            length=0,
            access_perms=MemoryPermissions.RX,
        ),
    )
    return LinkableSymbolStubInfo(asm_prefixes, segments)



import pefile as pefile

from ofrak.component.unpacker import Unpacker
from ofrak.core.code_region import CodeRegion
from ofrak.resource import Resource
from ofrak.core.pe.model import (
    Pe,
    Pe32OptionalHeader,
    Pe64POptionalHeader,
    PeMsDosHeader,
    PeFileHeader,
    PeOptionalHeader,
    PeDataDirectory,
    PeSectionHeader,
    PeSection,
    PeSectionFlag,
)
from ofrak_type.range import Range


class PeUnpacker(Unpacker[None]):
    id = b"PeUnpacker"
    targets = (Pe,)
    children = (
        PeMsDosHeader,
        PeFileHeader,
        PeOptionalHeader,
        PeDataDirectory,
        PeSectionHeader,
        PeSection,
    )

    async def unpack(self, resource: Resource, config=None):
        """
        Unpack a PE file using the pefile library for the parsing itself, then
        create the appropriate OFRAK child resources for the PE file.
        """
        pe = pefile.PE(data=await resource.get_data())

        # MS-DOS header
        ms_dos_header = self.ms_dos_header_from_pefile(pe.DOS_HEADER)
        await resource.create_child_from_view(
            ms_dos_header,
            data_range=Range.from_size(pe.DOS_HEADER.get_file_offset(), pe.DOS_HEADER.sizeof()),
        )

        # File header
        file_header = self.file_header_from_pefile(pe.FILE_HEADER)
        await resource.create_child_from_view(
            file_header,
            data_range=Range.from_size(pe.FILE_HEADER.get_file_offset(), pe.FILE_HEADER.sizeof()),
        )

        # Optional header
        if pe.OPTIONAL_HEADER is not None:
            optional_header = self.optional_header_from_pefile(pe.OPTIONAL_HEADER)
            await resource.create_child_from_view(
                optional_header,
                data_range=Range.from_size(
                    pe.OPTIONAL_HEADER.get_file_offset(), pe.OPTIONAL_HEADER.sizeof()
                ),
            )

            # Data directories
            for pe_data_directory in pe.OPTIONAL_HEADER.DATA_DIRECTORY:
                data_directory = self.data_directory_from_pefile(pe_data_directory)
                await resource.create_child_from_view(
                    data_directory,
                    data_range=Range.from_size(
                        pe_data_directory.get_file_offset(), pe_data_directory.sizeof()
                    ),
                )

        # Sections
        for index, pe_section in enumerate(pe.sections):
            # Section header
            section_header = self.section_header_from_pefile(pe_section, index)
            await resource.create_child_from_view(
                section_header,
                data_range=Range.from_size(pe_section.get_file_offset(), pe_section.sizeof()),
            )
            # The section itself
            section = PeSection(
                section_index=index,
                name=section_header.name,
                virtual_address=section_header.m_virtual_address,
                size=section_header.m_virtual_size,
            )
            if pe_section.SizeOfRawData > 0:
                section_range = Range.from_size(
                    pe_section.PointerToRawData, pe_section.SizeOfRawData
                )
            else:
                section_range = None

            section_r = await resource.create_child_from_view(
                section,
                data_range=section_range,
            )

            if section_header.has_flag(PeSectionFlag.IMAGE_SCN_CNT_CODE):
                section_r.add_tag(CodeRegion)

    @staticmethod
    def ms_dos_header_from_pefile(dos_header: pefile.Structure) -> PeMsDosHeader:
        return PeMsDosHeader(
            e_magic=dos_header.e_magic,
            e_cblp=dos_header.e_cblp,
            e_cp=dos_header.e_cp,
            e_crlc=dos_header.e_crlc,
            e_cparhdr=dos_header.e_cparhdr,
            e_minalloc=dos_header.e_minalloc,
            e_maxalloc=dos_header.e_maxalloc,
            e_ss=dos_header.e_ss,
            e_sp=dos_header.e_sp,
            e_csum=dos_header.e_csum,
            e_ip=dos_header.e_ip,
            e_cs=dos_header.e_cs,
            e_lfarlc=dos_header.e_lfarlc,
            e_ovno=dos_header.e_ovno,
            e_res=dos_header.e_res,
            e_oemid=dos_header.e_oemid,
            e_oeminfo=dos_header.e_oeminfo,
            e_res2=dos_header.e_res2,
            e_lfanew=dos_header.e_lfanew,
        )

    @staticmethod
    def file_header_from_pefile(file_header: pefile.Structure) -> PeFileHeader:
        return PeFileHeader(
            machine=file_header.Machine,
            number_of_sections=file_header.NumberOfSections,
            time_date_stamp=file_header.TimeDateStamp,
            pointer_to_symbol_table=file_header.PointerToSymbolTable,
            number_of_symbols=file_header.NumberOfSymbols,
            size_of_optional_header=file_header.SizeOfOptionalHeader,
            characteristics=file_header.Characteristics,
        )

    @staticmethod
    def optional_header_from_pefile(optional_header: pefile.Structure) -> PeOptionalHeader:
        if optional_header.name == "IMAGE_OPTIONAL_HEADER64":
            return Pe64POptionalHeader(
                magic=optional_header.Magic,
                major_linker_version=optional_header.MajorLinkerVersion,
                minor_linker_version=optional_header.MinorLinkerVersion,
                size_of_code=optional_header.SizeOfCode,
                size_of_initialized_data=optional_header.SizeOfInitializedData,
                size_of_uninitialized_data=optional_header.SizeOfUninitializedData,
                address_of_entry_point=optional_header.AddressOfEntryPoint,
                base_of_code=optional_header.BaseOfCode,
                image_base=optional_header.ImageBase,
                section_alignment=optional_header.SectionAlignment,
                file_alignment=optional_header.FileAlignment,
                major_operating_system_version=optional_header.MajorOperatingSystemVersion,
                minor_operating_system_version=optional_header.MinorOperatingSystemVersion,
                major_image_version=optional_header.MajorImageVersion,
                minor_image_version=optional_header.MinorImageVersion,
                major_subsystem_version=optional_header.MajorSubsystemVersion,
                minor_subsystem_version=optional_header.MinorSubsystemVersion,
                size_of_image=optional_header.SizeOfImage,
                size_of_headers=optional_header.SizeOfHeaders,
                checksum=optional_header.CheckSum,
                subsystem=optional_header.Subsystem,
                dll_characteristics=optional_header.DllCharacteristics,
                size_of_stack_reserve=optional_header.SizeOfStackReserve,
                size_of_stack_commit=optional_header.SizeOfStackCommit,
                size_of_heap_reserve=optional_header.SizeOfHeapReserve,
                size_of_heap_commit=optional_header.SizeOfHeapCommit,
                loader_flags=optional_header.LoaderFlags,
                number_of_rva_and_sizes=optional_header.NumberOfRvaAndSizes,
            )
        elif optional_header.name == "IMAGE_OPTIONAL_HEADER":
            return Pe32OptionalHeader(
                magic=optional_header.Magic,
                major_linker_version=optional_header.MajorLinkerVersion,
                minor_linker_version=optional_header.MinorLinkerVersion,
                size_of_code=optional_header.SizeOfCode,
                size_of_initialized_data=optional_header.SizeOfInitializedData,
                size_of_uninitialized_data=optional_header.SizeOfUninitializedData,
                address_of_entry_point=optional_header.AddressOfEntryPoint,
                base_of_code=optional_header.BaseOfCode,
                base_of_data=optional_header.BaseOfData,
                image_base=optional_header.ImageBase,
                section_alignment=optional_header.SectionAlignment,
                file_alignment=optional_header.FileAlignment,
                major_operating_system_version=optional_header.MajorOperatingSystemVersion,
                minor_operating_system_version=optional_header.MinorOperatingSystemVersion,
                major_image_version=optional_header.MajorImageVersion,
                minor_image_version=optional_header.MinorImageVersion,
                major_subsystem_version=optional_header.MajorSubsystemVersion,
                minor_subsystem_version=optional_header.MinorSubsystemVersion,
                size_of_image=optional_header.SizeOfImage,
                size_of_headers=optional_header.SizeOfHeaders,
                checksum=optional_header.CheckSum,
                subsystem=optional_header.Subsystem,
                dll_characteristics=optional_header.DllCharacteristics,
                size_of_stack_reserve=optional_header.SizeOfStackReserve,
                size_of_stack_commit=optional_header.SizeOfStackCommit,
                size_of_heap_reserve=optional_header.SizeOfHeapReserve,
                size_of_heap_commit=optional_header.SizeOfHeapCommit,
                loader_flags=optional_header.LoaderFlags,
                number_of_rva_and_sizes=optional_header.NumberOfRvaAndSizes,
            )
        else:
            raise ValueError("Optional header type not recognized")

    @staticmethod
    def data_directory_from_pefile(data_directory: pefile.Structure) -> PeDataDirectory:
        return PeDataDirectory(
            name=data_directory.name,
            virtual_address=data_directory.VirtualAddress,
            size=data_directory.Size,
        )

    @staticmethod
    def section_header_from_pefile(section_header: pefile.Structure, index: int) -> PeSectionHeader:
        return PeSectionHeader(
            m_name=section_header.Name,
            m_virtual_size=section_header.Misc_VirtualSize,
            m_virtual_address=section_header.VirtualAddress,
            m_size_of_raw_data=section_header.SizeOfRawData,
            m_pointer_to_raw_data=section_header.PointerToRawData,
            m_pointer_to_relocations=section_header.PointerToRelocations,
            m_pointer_to_linenumbers=section_header.PointerToLinenumbers,
            m_number_of_relocations=section_header.NumberOfRelocations,
            m_number_of_linenumbers=section_header.NumberOfLinenumbers,
            m_characteristics=section_header.Characteristics,
            section_index=index,
        )



from dataclasses import dataclass


from ofrak.model.resource_model import index
from ofrak.resource_view import ResourceView


@dataclass
class Addressable(ResourceView):
    """
    A resource with a virtual address.

    :ivar virtual_address: the virtual address
    """

    virtual_address: int

    @index
    def VirtualAddress(self) -> int:
        return self.virtual_address



from abc import ABC, abstractmethod

from ofrak.component.unpacker import Unpacker
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.memory_region import MemoryRegion
from ofrak.resource import Resource


class CodeRegion(MemoryRegion):
    """
    A memory region within a [program][ofrak.core.program.Program] that contains executable code.
    """


class CodeRegionUnpacker(Unpacker[None], ABC):
    """
    Unpack a [code region][ofrak.core.code_region.CodeRegion] into
    [complex blocks][ofrak.core.complex_block.ComplexBlock].
    """

    targets = (CodeRegion,)
    children = (ComplexBlock,)
    id = b"CodeRegionUnpacker"

    @abstractmethod
    async def unpack(self, resource: Resource, config=None):
        """
        Unpack a code region, extracting all of the complex blocks within it.

        The bounds of these complex blocks should include any trailing data literals which are
        considered part of the complex block only if there are only data references to them from
        within that complex block.

        :param resource: the code region resource
        :param config:

        :raises ComplexBlockStructureError: if the unpacker tries to define a complex block which
        did not match our expectations of complex block structure
        """
        raise NotImplementedError()



from collections import defaultdict
from typing import Any, Dict, cast

from itertools import product

from ofrak.service.serialization.pjson import PJSONSerializationService
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.service_i import SerializationServiceInterface


def _short_string_generator():
    """Generator of short JSON-compatible strings of increasing size."""
    disallowed_characters = [0x22, 0x5C, 0x2F, 0x62, 0x66, 0x6E, 0x72, 0x74, 0x75]
    alphabet = [chr(i) for i in range(0x20, 0x7E) if i not in disallowed_characters]
    size = 1
    while True:
        for letters in product(*([alphabet] * size)):
            yield "".join(letters)
        size += 1


class StashedPJSONSerializationService(SerializationServiceInterface):
    """
    Serialization service producing a more compact encoding than `PJSONSerializationService`, at the expense of time
    performance, by mapping each string in the PJSON to a shorter one, and including the mapping at the beginning of
    the serialized form.
    """

    def __init__(self, serializer: PJSONSerializationService):
        self._serializer = serializer

    def to_pjson(self, obj: Any, type_hint: Any) -> PJSONType:
        pjson = self._serializer.to_pjson(obj, type_hint)
        return self._shorten(pjson)

    def from_pjson(self, pjson_obj: PJSONType, type_hint: Any) -> Any:
        restored_pjson = self._restore(pjson_obj)
        return self._serializer.from_pjson(restored_pjson, type_hint)

    def to_json(self, obj: Any, type_hint: Any) -> str:
        return self._serializer.dumps(self.to_pjson(obj, type_hint))

    def from_json(self, json_obj: str, type_hint: Any) -> Any:
        return self.from_pjson(self._serializer.loads(json_obj), type_hint)

    def _shorten(self, pjson: PJSONType) -> Dict[str, PJSONType]:
        strings_long_to_short: Dict[str, str] = defaultdict(_short_string_generator().__next__)
        shortened_pjson = self._convert(pjson, strings_long_to_short)
        return {
            "string_mapping": {short: long for long, short in strings_long_to_short.items()},
            "shortened_pjson": shortened_pjson,
        }

    def _restore(self, shortened_pjson: PJSONType) -> PJSONType:
        pjson_obj = cast(Dict[str, PJSONType], shortened_pjson)
        strings_short_to_long = pjson_obj["string_mapping"]
        strings_short_to_long = cast(Dict[str, str], strings_short_to_long)
        shortened_pjson = pjson_obj["shortened_pjson"]
        return self._convert(shortened_pjson, strings_short_to_long)

    def _convert(
        self,
        pjson: PJSONType,
        string_mapping: Dict[str, str],
    ) -> PJSONType:
        """
        Perform the string replacement according to `string_mapping`.

        This method is used for both encoding and decoding, with opposite `string_mapping` dictionaries.
        """
        if isinstance(pjson, str):
            return string_mapping[pjson]
        elif isinstance(pjson, list):
            return [self._convert(item, string_mapping) for item in pjson]
        elif isinstance(pjson, tuple):
            return tuple(self._convert(item, string_mapping) for item in pjson)
        elif isinstance(pjson, dict):
            return {
                cast(str, self._convert(key, string_mapping)): self._convert(value, string_mapping)
                for key, value in pjson.items()
            }
        else:
            return pjson



import binascii
import json
import os.path
import uuid
from git import Repo
from dataclasses import dataclass
from typing import Dict, List, Optional

from ofrak.core.run_script_modifier import RunScriptModifier, RunScriptModifierConfig

from ofrak.resource import Resource

from ofrak.ofrak_context import OFRAKContext


@dataclass
class _OfrakProjectBinary:
    associated_scripts: List[str]
    init_script: Optional[str]


class OfrakProject:
    """
    An OFRAK 'project'

    """

    def __init__(
        self,
        path: str,
        name: str,
        project_id: bytes,
        binaries: Dict[str, _OfrakProjectBinary],
        scripts: List[str],
    ):
        self.path: str = path
        self.name: str = name
        self.project_id: bytes = project_id
        self.binaries: Dict[str, _OfrakProjectBinary] = binaries
        self.scripts: List[str] = scripts
        self.session_id = uuid.uuid4().bytes

    @property
    def metadata_path(self):
        return os.path.join(self.path, "metadata.json")

    @property
    def readme_path(self):
        return os.path.join(self.path, "README.md")

    @property
    def trashed_binaries(self):
        return os.path.join(self.path, ".Trash", "binaries")

    @property
    def trashed_scripts(self):
        return os.path.join(self.path, ".Trash", "binaries")

    @staticmethod
    def create(name: str, path: str) -> "OfrakProject":
        new_project = OfrakProject(
            path,
            name,
            uuid.uuid4().bytes,
            {},
            [],
        )

        os.makedirs(os.path.join(path, "scripts"), exist_ok=True)
        os.makedirs(os.path.join(path, "binaries"), exist_ok=True)

        new_project.write_metadata_to_disk()

        return new_project

    @staticmethod
    def clone_from_git(url: str, path: str) -> "OfrakProject":
        repo = Repo.clone_from(url, path)

        return OfrakProject.init_from_path(path)

    @staticmethod
    def init_from_path(path: str) -> "OfrakProject":
        """

        Assume path points to a directory with the following structure:
        (top-level directory)
        |-metadata.json
        |-README.md
        |--binaries
        |   |-binary1.bin
        |   | ...
        |--scripts
            |-script1.py
            | ...

        :param path:
        :return:
        """
        if not os.path.exists(path):
            raise ValueError(f"{path} does not exist")
        if not os.path.isdir(path):
            raise ValueError(f"{path} is not a directory")

        metadata_path = os.path.join(path, "metadata.json")
        binaries_path = os.path.join(path, "binaries")
        scripts_path = os.path.join(path, "scripts")

        if not all(
            [
                os.path.exists(metadata_path),
                os.path.exists(binaries_path),
                os.path.isdir(binaries_path),
                os.path.exists(scripts_path),
                os.path.isdir(scripts_path),
            ]
        ):
            raise ValueError(f"{path} has invalid structure to be an Project")

        with open(metadata_path) as f:
            raw_metadata = json.load(f)

        scripts = [script["name"] for script in raw_metadata["scripts"]]

        binaries = {}

        for binary_name, info in raw_metadata["binaries"].items():
            binaries[binary_name] = _OfrakProjectBinary(
                info["associated_scripts"], info.get("init_script")
            )
        name = raw_metadata["name"]
        project_id = binascii.unhexlify(raw_metadata["project_id"])
        project = OfrakProject(
            path,
            name,
            project_id,
            binaries,
            scripts,
        )

        project._restore_trashed_files()
        return project

    def script_path(self, script_name, check: bool = True) -> str:
        if check and script_name not in self.scripts:
            raise ValueError(f"Script {script_name} is not a script in this Project")
        p = os.path.join(self.path, "scripts", script_name)
        if check and not os.path.exists(p):
            raise ValueError(
                f"Script {script_name} is known to this Project but is not on disk "
                f"(looked at {p})"
            )
        return p

    def binary_path(self, binary_name, check: bool = True) -> str:
        if check and binary_name not in self.binaries:
            raise ValueError(f"Binary {binary_name} is not a binary in this Project")
        p = os.path.join(self.path, "binaries", binary_name)
        if check and not os.path.exists(p):
            raise ValueError(
                f"Binary {binary_name} is known to this Project but is not on disk "
                f"(looked at {p})"
            )
        return p

    async def init_project_binary(
        self, binary_name: str, ofrak_context: OFRAKContext, script_name: Optional[str] = None
    ) -> Resource:
        if script_name is None:
            binary_metadata = self._get_binary(binary_name)
            script_name = binary_metadata.init_script

        resource = await ofrak_context.create_root_resource_from_file(self.binary_path(binary_name))

        if script_name is not None:
            with open(self.script_path(script_name)) as f:
                code = f.read()
            await resource.run(RunScriptModifier, RunScriptModifierConfig(code))

        return resource

    def get_script_body(self, script_name: str) -> str:
        if script_name not in self.scripts:
            raise ValueError(f"Script {script_name} is not a script in this Project")
        with open(self.script_path(script_name)) as f:
            code = f.read()
        return code

    def write_metadata_to_disk(self):
        metadata = {
            "name": self.name,
            "project_id": self.project_id.hex(),
            "scripts": [
                {
                    "name": script_name,
                }
                for script_name in self.scripts
            ],
            "binaries": {
                binary_name: {
                    "init_script": binary_info.init_script,
                    "associated_scripts": binary_info.associated_scripts,
                }
                for binary_name, binary_info in self.binaries.items()
            },
        }
        with open(self.metadata_path, "w") as f:
            json.dump(metadata, f)

    def get_current_metadata(self):
        return {
            "name": self.name,
            "project_id": self.project_id.hex(),
            "session_id": self.session_id.hex(),
            "scripts": [
                {
                    "name": script_name,
                }
                for script_name in self.scripts
            ],
            "binaries": {
                binary_name: {
                    "init_script": binary_info.init_script,
                    "associated_scripts": binary_info.associated_scripts,
                }
                for binary_name, binary_info in self.binaries.items()
            },
        }

    def add_binary(self, name: str, contents: bytes):
        self.binaries[name] = _OfrakProjectBinary([], None)
        os.makedirs(os.path.join(self.path, "binaries"), exist_ok=True)
        with open(self.binary_path(name, check=False), "wb+") as f:
            f.write(contents)

    def add_script(self, name: str, script_contents: str):
        self.scripts.append(name)
        os.makedirs(os.path.join(self.path, "scripts"), exist_ok=True)
        with open(self.script_path(name, check=False), "w+") as f:
            f.write(script_contents)

    def delete_binary(self, name: str):
        os.makedirs(self.trashed_binaries, exist_ok=True)
        os.rename(
            self.binary_path(name),
            os.path.join(self.trashed_binaries, name),
        )
        self.binaries.pop(name)

    def delete_script(self, name: str):
        os.makedirs(self.trashed_scripts, exist_ok=True)
        os.rename(
            self.script_path(name),
            os.path.join(self.trashed_scripts, name),
        )
        self.scripts.remove(name)
        for binary_metadata in self.binaries.values():
            try:
                binary_metadata.associated_scripts.remove(name)
            except ValueError:
                pass
            if binary_metadata.init_script == name:
                binary_metadata.init_script = None

    def reset_project(self):
        path = self.path
        if not os.path.exists(path):
            raise ValueError(f"{path} does not exist")
        if not os.path.isdir(path):
            raise ValueError(f"{path} is not a directory")

        binaries_path = os.path.join(path, "binaries")
        scripts_path = os.path.join(path, "scripts")

        if not all(
            [
                os.path.exists(self.metadata_path),
                os.path.exists(binaries_path),
                os.path.isdir(binaries_path),
                os.path.exists(scripts_path),
                os.path.isdir(scripts_path),
            ]
        ):
            raise ValueError(f"{path} has invalid structure to be an Project")

        with open(self.metadata_path) as f:
            raw_metadata = json.load(f)

        self.scripts = [script["name"] for script in raw_metadata["scripts"]]
        self.binaries = {}

        for binaryName, info in raw_metadata["binaries"].items():
            self.binaries[binaryName] = _OfrakProjectBinary(
                info["associated_scripts"], info.get("init_script")
            )

        self.name = raw_metadata["name"]
        self.project_id = binascii.unhexlify(raw_metadata["project_id"])

        self._restore_trashed_files()

    def _restore_trashed_files(self):
        for script in self.scripts:
            try:
                self.script_path(script)
            except ValueError:
                try:
                    os.rename(
                        os.path.join(self.trashed_scripts, script),
                        self.script_path(script, check=False),
                    )
                except OSError:
                    raise AttributeError(
                        f"Trying to restore script {script} but the file is missing from .Trash"
                    )

        for binary in self.binaries.keys():
            try:
                self.binary_path(binary)
            except ValueError:
                try:
                    os.rename(
                        os.path.join(self.trashed_binaries, binary),
                        self.binary_path(binary, check=False),
                    )
                except OSError:
                    raise AttributeError(
                        f"Trying to restore binary {binary} but the file is missing from .Trash"
                    )

    def _get_binary(self, name):
        if not name in self.binaries.keys():
            raise ValueError(f"Binary with the name {name} does not exist")
        return self.binaries[name]



import io
import struct
import zlib
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List

from ofrak.component.identifier import Identifier
from ofrak.component.analyzer import Analyzer
from ofrak.component.modifier import Modifier
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker, UnpackerError
from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import File
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak.resource_view import ResourceView
from ofrak.service.resource_service_i import ResourceFilter
from ofrak_io.deserializer import BinaryDeserializer
from ofrak_type.endianness import Endianness
from ofrak_type.range import Range

OPENWRT_TRX_MAGIC_BYTES = b"HDR0"
(OPENWRT_TRX_MAGIC_START,) = struct.unpack("<I", OPENWRT_TRX_MAGIC_BYTES)
OPENWRT_TRXV1_HEADER_LEN = 28
OPENWRT_TRXV2_HEADER_LEN = 32
OPENWRT_TRX_MARK = struct.pack(">I", 0xDEADC0DE)


#####################
#       Enums       #
#####################
class OpenWrtTrxVersion(Enum):
    VERSION1 = 1
    VERSION2 = 2


#####################
#     RESOURCES     #
#####################
@dataclass
class OpenWrtTrxHeader(ResourceView):
    """
    OpenWrt trx header
    Information from <https://openwrt.org/docs/techref/headers>
    ```
    TRX v1
     0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     +---------------------------------------------------------------+
     |                     magic number ('HDR0')                     |
     +---------------------------------------------------------------+
     |                  length (header size + data)                  |
     +---------------+---------------+-------------------------------+
     |                       32-bit CRC value                        |
     +---------------+---------------+-------------------------------+
     |           TRX flags           |          TRX version          |
     +-------------------------------+-------------------------------+
     |                      Partition offset[0]                      |
     +---------------------------------------------------------------+
     |                      Partition offset[1]                      |
     +---------------------------------------------------------------+
     |                      Partition offset[2]                      |
     +---------------------------------------------------------------+

    TRX v2
      0                   1                   2                   3
      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
     +---------------------------------------------------------------+
     |                     magic number ('HDR0')                     |
     +---------------------------------------------------------------+
     |                  length (header size + data)                  |
     +---------------+---------------+-------------------------------+
     |                       32-bit CRC value                        |
     +---------------+---------------+-------------------------------+
     |           TRX flags           |          TRX version          |
     +-------------------------------+-------------------------------+
     |                      Partition offset[0]                      |
     +---------------------------------------------------------------+
     |                      Partition offset[1]                      |
     +---------------------------------------------------------------+
     |                      Partition offset[2]                      |
     +---------------------------------------------------------------+
     |                      Partition offset[3]                      |
     +---------------------------------------------------------------+
    ```
    """

    trx_magic: int
    trx_length: int
    trx_crc: int
    trx_flags: int
    trx_version: int
    trx_partition_offsets: List[int]

    def get_version(self) -> OpenWrtTrxVersion:
        return OpenWrtTrxVersion(self.trx_version)

    def get_header_length(self) -> int:
        if self.get_version() == OpenWrtTrxVersion.VERSION1:
            return OPENWRT_TRXV1_HEADER_LEN
        elif self.get_version() == OpenWrtTrxVersion.VERSION2:
            return OPENWRT_TRXV2_HEADER_LEN
        else:
            raise ValueError(f"Unknown OpenWrt TRX version: {self.trx_version}")


@dataclass
class OpenWrtTrx(GenericBinary):
    """
    OpenWrtTrx is a TRX binary update for OpenWrt firmware.

    TRX binaries are used for upgrading the firmware of devices already running OpenWrt firmware

    The OpenWrtTrx consists of 1 OpenWrtTrxHeader and 3 or 4 partition(s).
    """

    async def get_header(self) -> OpenWrtTrxHeader:
        return await self.resource.get_only_child_as_view(
            OpenWrtTrxHeader, ResourceFilter.with_tags(OpenWrtTrxHeader)
        )


####################
#    IDENTIFIER    #
####################


class OpenWrtIdentifier(Identifier[None]):
    targets = (
        File,
        GenericBinary,
    )

    async def identify(self, resource: Resource, config=None) -> None:
        trx_magic = await resource.get_data(range=Range(0, 4))
        if trx_magic == OPENWRT_TRX_MAGIC_BYTES:
            resource.add_tag(OpenWrtTrx)


####################
#    UNPACKERS     #
####################
class OpenWrtTrxUnpacker(Unpacker[None]):
    """
    Unpack an OpenWrtTrx firmware file into its partitions.

    The header has a mapped data range, whereas all partitions are unmapped data.
    """

    id = b"OpenWrtTrxUnpacker"
    targets = (OpenWrtTrx,)
    children = (
        OpenWrtTrxHeader,
        GenericBinary,
    )

    async def unpack(self, resource: Resource, config=None):
        data = await resource.get_data()
        # Peek into TRX version to know how big the header is
        trx_version = OpenWrtTrxVersion(struct.unpack("<H", data[14:16])[0])

        if trx_version not in [OpenWrtTrxVersion.VERSION1, OpenWrtTrxVersion.VERSION2]:
            raise UnpackerError(f"Unknown OpenWrt TRX version: {trx_version}")

        header_len = (
            OPENWRT_TRXV1_HEADER_LEN
            if trx_version == OpenWrtTrxVersion.VERSION1
            else OPENWRT_TRXV2_HEADER_LEN
        )
        trx_header_r = await resource.create_child(
            tags=(OpenWrtTrxHeader,), data_range=Range(0, header_len)
        )

        trx_header = await trx_header_r.view_as(OpenWrtTrxHeader)
        partition_offsets = trx_header.trx_partition_offsets

        for i, offset in enumerate(partition_offsets):
            if offset == 0:
                break

            next_offset = (
                partition_offsets[i + 1] if i < (len(partition_offsets) - 1) else len(data)
            )
            partition = data[offset:next_offset]

            child = await resource.create_child(
                tags=(GenericBinary,), data_range=Range(offset, next_offset)
            )
            if OPENWRT_TRX_MARK in partition:
                partition = partition[: partition.index(OPENWRT_TRX_MARK)]
                await child.create_child(
                    tags=(GenericBinary,), data_range=Range.from_size(0, len(partition))
                )


#####################
#     ANALYZERS     #
#####################
class OpenWrtTrxHeaderAttributesAnalyzer(Analyzer[None, OpenWrtTrxHeader]):
    """
    Analyze the OpenWrtTrxHeader of a OpenWrtTrx firmware file.
    """

    targets = (OpenWrtTrxHeader,)
    outputs = (OpenWrtTrxHeader,)

    async def analyze(self, resource: Resource, config=None) -> OpenWrtTrxHeader:
        tmp = await resource.get_data()
        deserializer = BinaryDeserializer(
            io.BytesIO(tmp),
            endianness=Endianness.LITTLE_ENDIAN,
            word_size=4,
        )
        deserialized = deserializer.unpack_multiple("IIIHH")
        (
            trx_magic,
            trx_length,
            trx_crc,
            trx_flags,
            trx_version,
        ) = deserialized
        assert trx_magic == OPENWRT_TRX_MAGIC_START

        if trx_version not in [OpenWrtTrxVersion.VERSION1.value, OpenWrtTrxVersion.VERSION2.value]:
            raise ValueError(f"Unknown OpenWrt TRX version: {trx_version}")

        max_num_partitions = 3 if trx_version == OpenWrtTrxVersion.VERSION1.value else 4
        trx_partition_offsets = []

        for _ in range(max_num_partitions):
            offset = deserializer.unpack_uint()

            if offset == 0:
                break

            trx_partition_offsets.append(offset)

        header = OpenWrtTrxHeader(
            trx_magic, trx_length, trx_crc, trx_flags, trx_version, trx_partition_offsets
        )

        return header


####################
#    MODIFIERS     #
####################
@dataclass
class OpenWrtTrxHeaderModifierConfig(ComponentConfig):
    """
    Modifier config for a OpenWrtTrxHeader.

    The following field is not modifiable:
    - Image Header Magic Number (a constant): `trx_magic`

    The following fields are modifiable, but will be overwritten by the packer:
    - Image Data Size: `trx_length`
    - Image CRC Checksum: `trx_crc`
    - LZMA Loader offset: `trx_loader_offset`
    - Kernel offset: `trx_kernel_offset`
    - Filesystem offset: `trx_rootfs_offset`
    - BinHeader offset, applicable for TRX Version 2: `trx_binheader_offset`
    """

    trx_length: Optional[int] = None
    trx_crc: Optional[int] = None
    trx_flags: Optional[int] = None
    trx_version: Optional[int] = None
    trx_partition_offsets: Optional[List[int]] = None


class OpenWrtTrxHeaderModifier(Modifier[OpenWrtTrxHeaderModifierConfig]):
    """
    Modify a OpenWrtTrxHeader according to a given modifier config.
    """

    targets = (OpenWrtTrxHeader,)

    async def modify(self, resource: Resource, config: OpenWrtTrxHeaderModifierConfig) -> None:
        original_attributes = await resource.analyze(AttributesType[OpenWrtTrxHeader])
        new_attributes = ResourceAttributes.replace_updated(original_attributes, config)
        serialized_header = await OpenWrtTrxHeaderModifier.serialize(new_attributes)
        header_v = await resource.view_as(OpenWrtTrxHeader)
        resource.queue_patch(
            Range.from_size(0, header_v.get_header_length()),
            serialized_header,
        )
        resource.add_attributes(new_attributes)

    @staticmethod
    async def serialize(
        updated_attributes: AttributesType[OpenWrtTrxHeader],
    ) -> bytes:
        """
        Serialize `updated_attributes` into bytes. This method doesn't perform any check or compute
        any CRC.
        """
        output = struct.pack(
            "<IIIHH",
            OPENWRT_TRX_MAGIC_START,
            updated_attributes.trx_length,
            updated_attributes.trx_crc,
            updated_attributes.trx_flags,
            updated_attributes.trx_version,
        )
        if updated_attributes.trx_version == OpenWrtTrxVersion.VERSION1.value:
            num_offsets = 3
        elif updated_attributes.trx_version == OpenWrtTrxVersion.VERSION2.value:
            num_offsets = 4
        else:
            raise ValueError()
        for offset in updated_attributes.trx_partition_offsets:
            output += struct.pack("<I", offset)
        for i in range(num_offsets - len(updated_attributes.trx_partition_offsets)):
            output += struct.pack("<I", 0)
        return output


####################
#     PACKERS      #
####################
class OpenWrtTrxPacker(Packer[None]):
    """
    Pack an OpenWrtTrx firmware file.

    It consolidates the OpenWrtTrxHeader and all partition instances into a single binary, updating
    the CRC checksum, data size, and partition offsets in the header.
    """

    id = b"OpenWrtTrxPacker"
    targets = (OpenWrtTrx,)

    async def pack(self, resource: Resource, config=None):
        openwrt_v = await resource.view_as(OpenWrtTrx)
        header = await openwrt_v.get_header()
        children_by_offset = sorted(
            [
                (await child.get_data_range_within_root(), child)
                for child in await resource.get_children()
                if not child.has_tag(OpenWrtTrxHeader)
            ],
            key=lambda x: x[0].start,
        )
        repacked_data_l = [await child.get_data() for _, child in children_by_offset]
        repacked_data_b = b"".join(repacked_data_l)
        trx_length = header.get_header_length() + len(repacked_data_b)

        offsets = [r.start for r, _ in children_by_offset]
        header_config = OpenWrtTrxHeaderModifierConfig(
            trx_length=trx_length, trx_partition_offsets=offsets
        )

        await header.resource.run(OpenWrtTrxHeaderModifier, header_config)

        header_data = await header.resource.get_data()
        data_to_crc = header_data[12:] + repacked_data_b
        header_config = OpenWrtTrxHeaderModifierConfig(
            trx_crc=openwrt_crc32(data_to_crc),
        )
        await header.resource.run(OpenWrtTrxHeaderModifier, header_config)
        original_size = await resource.get_data_length()
        resource.queue_patch(Range(header.get_header_length(), original_size), repacked_data_b)


def openwrt_crc32(data: bytes) -> int:
    """
    Calculate CRC32 a-la OpenWrt. Original implementation:
    <https://git.archive.openwrt.org/?p=14.07/openwrt.git;a=blob;f=tools/firmware-utils/src/trx.c>

    Implements CRC-32 Ethernet which requires XOR'ing the zlib.crc32 result with 0xFFFFFFFF
    """
    return (zlib.crc32(data) & 0xFFFFFFFF) ^ 0xFFFFFFFF



import asyncio
import logging
import os
import tempfile
from dataclasses import dataclass
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import File, Folder, FilesystemRoot, SpecialFileType
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier

from ofrak.model.component_model import ComponentExternalTool
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)

SEVEN_ZIP = ComponentExternalTool("7zz", "https://www.7-zip.org", "--help", brew_package="sevenzip")


@dataclass
class SevenZFilesystem(GenericBinary, FilesystemRoot):
    """
    Filesystem stored in a 7z archive.
    """


class SevenZUnpacker(Unpacker[None]):
    """
    Unpack (decompress) a 7z file.
    """

    targets = (SevenZFilesystem,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (SEVEN_ZIP,)

    async def unpack(self, resource: Resource, config=None):
        seven_zip_v = await resource.view_as(SevenZFilesystem)
        resource_data = await seven_zip_v.resource.get_data()
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_file.write(resource_data)
            temp_file.flush()
            with tempfile.TemporaryDirectory() as temp_flush_dir:
                cmd = [
                    "7zz",
                    "x",
                    f"-o{temp_flush_dir}",
                    temp_file.name,
                ]
                proc = await asyncio.create_subprocess_exec(
                    *cmd,
                )
                returncode = await proc.wait()
                if proc.returncode:
                    raise CalledProcessError(returncode=returncode, cmd=cmd)
                await seven_zip_v.initialize_from_disk(temp_flush_dir)


class SevenzPacker(Packer[None]):
    """
    Pack files into a compressed 7z archive.
    """

    targets = (SevenZFilesystem,)
    external_dependencies = (SEVEN_ZIP,)

    async def pack(self, resource: Resource, config=None):
        seven_zip_v: SevenZFilesystem = await resource.view_as(SevenZFilesystem)
        temp_flush_dir = await seven_zip_v.flush_to_disk()
        temp_flush_dir = os.path.join(temp_flush_dir, ".")
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_name = os.path.join(temp_dir, "temp.7z")
            cmd = [
                "7zz",
                "a",
                temp_name,
                temp_flush_dir,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)
            with open(temp_name, "rb") as f:
                new_data = f.read()
            # Passing in the original range effectively replaces the original data with the new data
            resource.queue_patch(Range(0, await resource.get_data_length()), new_data)


MagicMimeIdentifier.register(SevenZFilesystem, "application/x-7z-compressed")
MagicDescriptionIdentifier.register(SevenZFilesystem, lambda s: s.startswith("7-zip archive"))



from typing import Dict, Type, Sequence, Any

from ofrak import ResourceTag
from ofrak.model.resource_model import ResourceModel, ResourceAttributes
from ofrak.service.serialization.pjson_types import PJSONType
from ofrak.service.serialization.serializers.serializer_i import SerializerInterface


class ResourceModelSerializer(SerializerInterface):
    """
    Serialize and deserialize `ResourceModel` into `PJSONType`.

    There are two issues with ResourceModel:
    1) the type hint of ResourceModel.attributes = Dict[Type[RA], RA] with
    RA = TypeVar("RA", bound="ResourceAttributes")
    This is used to refer to ResourceAttributes in the type hints without creating circular imports.
    But the serializer needs to know the real class ResourceAttributes, not just its name.
    2) the type hint of ResourceModel.tags is SortedSet[ResourceTag], but the generic serializer doesn't know how to
    handle SortedSet (to do it right, we would also need to serialize the sorting function, not just the sequence).
    But tags can be given to ResourceModel.__init__ as simply any Sequence, so we use that type hint instead.
    """

    targets = (ResourceModel,)

    # Start from the type hints of ResourceModel, but modify them for attributes and tags
    usable_type_hints = ResourceModel.__annotations__
    usable_type_hints["attributes"] = Dict[Type[ResourceAttributes], ResourceAttributes]
    usable_type_hints["tags"] = Sequence[ResourceTag]

    def obj_to_pjson(self, obj: Any, _type_hint: Any) -> Dict[str, PJSONType]:
        result = {}
        for attr_name, type_hint in self.usable_type_hints.items():
            result[attr_name] = self._service.to_pjson(getattr(obj, attr_name), type_hint)

        # Manually add captions to the generated dictionary since they are a read-only property,
        # and they should only be serialized, never deserialized and passed to an object's __init__
        result["caption"] = obj.caption

        return result

    def pjson_to_obj(self, pjson_obj: Dict[str, PJSONType], _type_hint: Any) -> Any:
        deserialized_attrs = {
            attr_name: self._service.from_pjson(pjson_obj[attr_name], type_hint)
            for attr_name, type_hint in self.usable_type_hints.items()
        }
        return ResourceModel(**deserialized_attrs)



import logging
from itertools import tee
from typing import List

from ofrak import Modifier, Resource, OFRAKContext
from ofrak.core import (
    Elf,
    FreeSpace,
    ElfProgramHeaderModifier,
    ElfProgramHeaderModifierConfig,
    Allocatable,
    ElfProgramHeader,
    ElfProgramHeaderType,
    ElfUnpacker,
)
from ofrak_type import Range


LOGGER = logging.getLogger()


class ElfLoadAlignmentModifier(Modifier[None]):
    """
    Reclaim unused alignment bytes between adjacent PT_LOAD segment in an ELF and tag them as free space.

    Inspired by Silvio Cesare's padding infection technique, this modifier capitalizes on the fact that the alignment
    bytes that exist between PT_LOAD sections in ELF binaries. These alignment bytes are added to the preceding PT_LOAD
    segment.

    Use this modifier if these alignment bytes provide enough space for your planned binary modifications.

    For more information on Silvio's padding infection technique, see:
    - The [original paper](https://web.archive.org/web/20130820033847/http://vxheaven.org/lib/vsc01.html)
    - Chapter 4 of Ryan "elfmaster" O'Neill's "Learning Linux Binary Analysis".
    """

    targets = (Elf,)

    async def modify(self, resource: Resource, config=None) -> None:
        """
        Scan target ELF for unused alignment bytes between PT_LOAD segment, extend preceding segment to include these
        bytes, and tag them as FreeSpace.
        """
        elf: Elf = await resource.view_as(Elf)
        load_program_headers = list(await get_load_program_headers(elf))
        # First, get program headers sorted by file offset
        load_program_headers_by_offset = sorted(load_program_headers, key=lambda x: x.p_offset)
        found_space = False
        for first_program_header, second_program_header in _pairwise(
            load_program_headers_by_offset
        ):
            if first_program_header.p_filesz != first_program_header.p_memsz:
                # Avoid overwriting .bss in data segments or other weird things that change size
                continue
            first_header_range = Range.from_size(
                first_program_header.p_offset, first_program_header.p_filesz
            )
            second_header_range = Range.from_size(
                second_program_header.p_offset, second_program_header.p_filesz
            )
            potential_free_space = second_header_range.start - first_header_range.end
            if potential_free_space <= 0:
                # No free space exists on disk
                continue
            found_space = True
            LOGGER.info(
                f"Found {hex(potential_free_space)} of free space between {first_program_header}and"
                f"{second_program_header}"
            )
            trimmed_potential_free_space = await self._trim_potential_free_space_by_vaddr(
                first_program_header, load_program_headers_by_offset, potential_free_space
            )
            free_space = FreeSpace(
                first_program_header.p_vaddr + first_program_header.p_memsz,
                trimmed_potential_free_space,
                first_program_header.get_memory_permissions(),
            )
            LOGGER.info(f"Creating {free_space} in {resource}")
            await resource.create_child_from_view(
                free_space,
                data_range=Range.from_size(first_header_range.end, trimmed_potential_free_space),
            )
            await first_program_header.resource.run(
                ElfProgramHeaderModifier,
                ElfProgramHeaderModifierConfig(
                    p_filesz=first_program_header.p_filesz + trimmed_potential_free_space,
                    p_memsz=first_program_header.p_filesz + trimmed_potential_free_space,
                ),
            )
        if found_space:
            # Tag ELF as Allocatable so FreeSpace can be used
            resource.add_tag(Allocatable)

    @staticmethod
    async def _trim_potential_free_space_by_vaddr(
        first_program_header: ElfProgramHeader,
        load_program_headers_by_offset: List[ElfProgramHeader],
        offset_difference: int,
    ) -> int:
        """
        Check that the potential free space does not overlap with any existing PT_LOAD segments.
        If there is overlap, trim the potential free space.
        """
        vaddr_ranges = Range.merge_ranges(
            [
                Range.from_size(header.p_vaddr, header.p_memsz)
                for header in load_program_headers_by_offset
            ]
        )
        # Reduce size based on vaddr constraints
        potential_free_space_range = Range.from_size(
            first_program_header.p_vaddr + first_program_header.p_memsz, offset_difference
        )
        for vaddr_range in vaddr_ranges:
            if potential_free_space_range.overlaps(vaddr_range):
                potential_free_space_range = _crop_range(potential_free_space_range, vaddr_range)
        return potential_free_space_range.length()


async def get_load_program_headers(elf: Elf) -> List[ElfProgramHeader]:
    load_program_headers = list()
    for program_header in await elf.get_program_headers():
        if program_header.p_type == ElfProgramHeaderType.LOAD.value:
            load_program_headers.append(program_header)
    return load_program_headers


def _crop_range(range_1: Range, range_2: Range) -> Range:
    """
    Split range_1 around range_2 and return the split range whose start is the same as range_1.
    """
    splits = range_1.split(range_2)
    for s in splits:
        if s.start == range_1.start:
            return s
    raise ValueError(
        f"Error removing intersection between {range_1} and {range_2} from {range_1}. "
        f"This should be unreachable."
    )


def _pairwise(iterable):
    """
    A helper function waiting for itertools.pairwise from Python 3.10:
    https://docs.python.org/3/library/itertools.html#itertools.pairwise.

    Usage: `_pairwise('ABCDEFG') --> AB BC CD DE EF FG`
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


async def main(ofrak_context: OFRAKContext, file: str, output_file: str):
    root_resource = await ofrak_context.create_root_resource_from_file(file)
    await root_resource.identify()
    await root_resource.run(ElfUnpacker)
    await root_resource.run(ElfLoadAlignmentModifier)
    if root_resource.has_tag(Allocatable):
        allocatable = await root_resource.view_as(Allocatable)
        print(f"[+] Free space found by ElfLoadAlignmentFreeSpaceModifier: {allocatable}")
        for key, value in allocatable.free_space_ranges.items():
            print(f"[+] {key} free space: {hex(sum([v.length() for v in value]))} bytes")
        # print(f"[+] Fre space by permissions: {}")
        await root_resource.flush_to_disk(output_file)
        print(f"[+] Output file written to {output_file}")
    else:
        print("[+] No free space found")


if __name__ == "__main__":
    import argparse

    from ofrak import OFRAK

    parser = argparse.ArgumentParser()
    parser.add_argument("--file", "-f")
    parser.add_argument("--output-file", "-o")
    args = parser.parse_args()
    ofrak = OFRAK()
    ofrak.run(main, args.file, args.output_file)  # type: ignore



import dataclasses
import logging
import os
from dataclasses import dataclass
from typing import Dict, Iterable, Mapping, Optional, Set, Tuple

from ofrak.component.modifier import Modifier
from ofrak.core.addressable import Addressable
from ofrak.core.architecture import ProgramAttributes
from ofrak.core.binary import GenericBinary
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.label import LabeledAddress
from ofrak.core.patch_maker.linkable_symbol import LinkableSymbol
from ofrak.model.component_model import ComponentConfig
from ofrak.resource import Resource
from ofrak.service.resource_service_i import (
    ResourceFilter,
    ResourceAttributeValueFilter,
    ResourceAttributeValuesFilter,
)
from ofrak_patch_maker.model import BOM, PatchRegionConfig
from ofrak_patch_maker.toolchain.model import Segment
from ofrak_type import InstructionSet, InstructionSetMode, LinkableSymbolType, NotFoundError

LOGGER = logging.getLogger()


class SymbolExistsError(RuntimeError):
    pass


@dataclass
class LinkableBinary(GenericBinary):
    """
    A resource with accompanying metadata which provide a mapping
    between symbols, and their corresponding offset and type within the program.
    """

    async def get_only_symbol(
        self, *, name: Optional[str] = None, vaddr: Optional[int] = None
    ) -> LinkableSymbol:
        """
        Get exactly one LinkableSymbol from this LinkableBinary matching a given name,
        virtual address, or both. If values for both vaddr and name are specified, only finds
        symbols which match both vaddr AND name, not vaddr OR name.

        :param name: Name of the symbol to look for.
        :param vaddr: Virtual address of the symbol to look for.

        :return: The only LinkableSymbol with the given name and/or vaddr.

        :raises ValueError: if both `name` and `vaddr` arguments are None.
        :raises NotFoundError: if no symbol or more than one is found with name and/or vaddr.
        """
        attributes_filters = []
        if name is None and vaddr is None:
            raise ValueError("At least one of the arguments name, vaddr must not be None!")
        if vaddr is not None:
            attributes_filters.append(
                ResourceAttributeValueFilter(LinkableSymbol.VirtualAddress, vaddr)
            )
        if name is not None:
            attributes_filters.append(ResourceAttributeValueFilter(LinkableSymbol.Label, name))

        return await self.resource.get_only_descendant_as_view(
            LinkableSymbol,
            r_filter=ResourceFilter(
                tags=(LinkableSymbol,),
                attribute_filters=tuple(attributes_filters),
            ),
        )

    async def get_symbols(
        self, *, name: Optional[str] = None, vaddr: Optional[int] = None
    ) -> Iterable[LinkableSymbol]:
        """
        Get exactly all LinkableSymbols from this LinkableBinary matching a given name,
        virtual address, or both. If values for both vaddr and name are specified, only finds
        symbols which match both vaddr AND name, not vaddr OR name.

        :param name: Name of the symbols to look for.
        :param vaddr: Virtual address of the symbols to look for.

        :return: All LinkableSymbol with the given name and/or vaddr.

        :raises NotFoundError: if no symbols are found with name and/or vaddr.
        """
        attributes_filters = []
        if vaddr is not None:
            attributes_filters.append(
                ResourceAttributeValueFilter(LinkableSymbol.VirtualAddress, vaddr)
            )
        if name is not None:
            attributes_filters.append(ResourceAttributeValueFilter(LinkableSymbol.Label, name))

        return await self.resource.get_descendants_as_view(
            LinkableSymbol,
            r_filter=ResourceFilter(
                tags=(LinkableSymbol,),
                attribute_filters=tuple(attributes_filters) if attributes_filters else None,
            ),
        )

    async def define_linkable_symbols(
        self, proto_symbols: Dict[str, Tuple[int, LinkableSymbolType]]
    ):
        """
        From some basic info about symbols in this program, create a LinkableSymbol resource for
        each one and get any remaining needed info to do this. Usage is to pass a dictionary that
        defines each symbol like so:
        "symbol_name": (symbol_vaddr, symbol_type)

        :param proto_symbols: Mapping of each symbol name to its known vaddr and symbol type.

        :raises NotFoundError: if a ComplexBlock resource with the indicated vaddr does not exist
        for a provided FUNC symbol.
        """
        symbols = []
        for sym_name, (sym_vaddr, sym_type) in proto_symbols.items():
            mode = InstructionSetMode.NONE
            if sym_type is LinkableSymbolType.FUNC:
                try:
                    cb = await self.resource.get_only_descendant_as_view(
                        ComplexBlock,
                        r_filter=ResourceFilter(
                            tags=(ComplexBlock,),
                            attribute_filters=(
                                ResourceAttributeValueFilter(
                                    ComplexBlock.VirtualAddress, sym_vaddr
                                ),
                            ),
                        ),
                    )
                except NotFoundError:
                    raise NotFoundError(
                        f"No ComplexBlock resource exists at vaddr 0x"
                        f"{sym_vaddr:x}; cannot infer its mode."
                    )
                mode = await cb.get_mode()
            symbols.append(LinkableSymbol(sym_vaddr, sym_name, sym_type, mode))

        await self.resource.run(
            UpdateLinkableSymbolsModifier,
            UpdateLinkableSymbolsModifierConfig(tuple(symbols)),
        )

    async def define_linkable_symbols_from_patch(
        self,
        proto_symbols: Mapping[str, Tuple[int, LinkableSymbolType]],
        program_attributes: ProgramAttributes,
    ):
        """
        From some basic info about symbols returned from
        [get_bin_file_symbols][ofrak_patch_maker.toolchain.abstract.get_bin_file_symbols],
        create a LinkableSymbol resource for each one. Usage is to pass a dictionary that defines
        each symbol like so:
        "symbol_name": (symbol_vaddr, symbol_type)

        :param proto_symbols: Mapping of each symbol name to its known vaddr and symbol type.
        :param program_attributes: Attributes of the program being patched, assumed to be the same
        ISA as the patch files.
        """
        symbols = []
        for sym_name, (sym_vaddr, sym_type) in proto_symbols.items():
            mode = InstructionSetMode.NONE
            if sym_type is LinkableSymbolType.FUNC:
                if program_attributes.isa == InstructionSet.ARM and sym_vaddr & 0x01 == 1:
                    mode = InstructionSetMode.THUMB
                symbols.append(LinkableSymbol(sym_vaddr, sym_name, sym_type, mode))

        await self.resource.run(
            UpdateLinkableSymbolsModifier,
            UpdateLinkableSymbolsModifierConfig(tuple(symbols)),
        )

    # TODO: Un-stringify PatchMaker; OFRAK imports in PatchMaker results in circular Program imports
    async def make_linkable_bom(
        self,
        patch_maker: "PatchMaker",  # type: ignore
        build_tmp_dir: str,
        unresolved_symbols: Set[str],
    ) -> Tuple[BOM, PatchRegionConfig]:
        """
        Build a BOM with all the symbols known to SymbolizedBinary. This BOM can be used to build
        the FEM so that it has access to all those symbols as weak symbols. This BOM in practice
        enables linking against the target binary. Because they are weak symbols, the patch is
        free to redefine any of these symbols - for example, if we are patching in a new body for
        a function, it will not conflict with the existing symbol for that function in the target
        binary.

        We create stubs for symbols rather than treat them purely as symbols because the linker in
        some cases needs a little more information than the symbol alone can provide - ARM/Thumb
        interworking in particular pushed us in this direction. Creating stubs for all symbols is
        simpler than trying to decide whether we need to create a symbol or stub for each symbol
        based on its mode compared to the mode of something we are trying to inject.

        :param patch_maker: PatchMaker instance to use to build the BOM.
        :param build_tmp_dir: A temporary directory to use for the BOM files.
        :param unresolved_symbols: All symbols used in a patch BOM but defined elsewhere. May be
        defined in target binary or in an earlier patch.

        :return: Tuple consisting of a BOM and PatchConfig representing the weak symbol
        definitions for all LinkableSymbols in the binary, ready to be passed in the `boms`
        argument to PatchMaker.make_fem(...).
        """
        stubs: Dict[str, Tuple[Segment, ...]] = dict()
        for symbol in await self.get_symbols():
            if symbol.name in unresolved_symbols:
                stubs_file = os.path.join(build_tmp_dir, f"stub_{symbol.name}.as")
                stub_info = symbol.get_stub_info()
                stub_body = "\n".join(
                    stub_info.asm_prefixes + [f".global {symbol.name}", f"{symbol.name}:", ""]
                )

                with open(stubs_file, "w+") as f:
                    f.write(stub_body)
                stubs[stubs_file] = stub_info.segments

        if stubs:
            stubs_bom = patch_maker.make_bom(
                name="stubs",
                source_list=list(stubs.keys()),
                object_list=[],
                header_dirs=[],
            )
            stubs_object_segments = {
                stubs_bom.object_map[stubs_file].path: stub_segments
                for stubs_file, stub_segments in stubs.items()
            }

        else:
            empty_source = os.path.join(build_tmp_dir, "empty_source.c")
            with open(empty_source, "w") as f:
                pass

            stubs_bom = patch_maker.make_bom(
                name="stubs",
                source_list=[empty_source],
                object_list=[],
                header_dirs=[],
            )

            stubs_object_segments = {stubs_bom.object_map[empty_source].path: ()}

        return stubs_bom, PatchRegionConfig("stubs_segments", stubs_object_segments)


@dataclass
class UpdateLinkableSymbolsModifierConfig(ComponentConfig):
    updated_symbols: Tuple[LinkableSymbol, ...]

    def __post_init__(self):
        stripped_symbols = []
        for symbol in self.updated_symbols:
            stripped_symbol = dataclasses.replace(symbol)
            stripped_symbol._resource = None
            stripped_symbols.append(stripped_symbol)
        self.updated_symbols = tuple(stripped_symbols)


class UpdateLinkableSymbolsModifier(Modifier[UpdateLinkableSymbolsModifierConfig]):
    """
    Add or update the LinkableSymbols in a LinkableBinary.

    :raises ValueError: if not all the provided symbols have unique vaddrs.
    """

    targets = (LinkableBinary,)

    async def modify(self, resource: Resource, config: UpdateLinkableSymbolsModifierConfig) -> None:
        unhandled_symbols = {symbol.name: symbol for symbol in config.updated_symbols}
        unhandled_vaddrs: Dict[int, LinkableSymbol] = {}
        for symbol in config.updated_symbols:
            if symbol.virtual_address not in unhandled_vaddrs:
                unhandled_vaddrs[symbol.virtual_address] = symbol
            else:
                # Multiple symbols defined for an address (which is valid)
                # However it would be pointless to look for and rename an existing resource at that
                # address multiple times
                # So after the first one, create a new child for each symbol sharing an address
                await resource.create_child_from_view(symbol)

        # Overwrite existing ComplexBlock with new LinkableSymbols
        filter_for_cb_by_vaddr = ResourceFilter(
            tags=(ComplexBlock,),
            attribute_filters=(
                ResourceAttributeValuesFilter(
                    ComplexBlock.VirtualAddress,
                    tuple(
                        vaddr
                        for vaddr, symbol in unhandled_vaddrs.items()
                        if symbol.symbol_type is LinkableSymbolType.FUNC
                    ),
                ),
            ),
        )

        try:
            for existing_cb in await resource.get_descendants_as_view(
                ComplexBlock, r_filter=filter_for_cb_by_vaddr
            ):
                symbol = unhandled_vaddrs[existing_cb.virtual_address]
                existing_cb.resource.add_view(symbol)
                # Update the value of the ComplexBlock.name attribute
                existing_cb.resource.add_view(dataclasses.replace(existing_cb, name=symbol.name))
                unhandled_symbols.pop(symbol.name)
        except NotFoundError:
            LOGGER.debug("No existing ComplexBlocks found for the provided symbols, moving on")

        # Overwrite existing LabeledAddress with new LinkableSymbols
        filter_for_label_by_name = ResourceFilter(
            tags=(LabeledAddress,),
            attribute_filters=(
                ResourceAttributeValuesFilter(
                    LabeledAddress.Label, tuple(unhandled_symbols.keys())
                ),
            ),
        )
        try:
            for existing_label in await resource.get_descendants_as_view(
                LabeledAddress,
                r_filter=filter_for_label_by_name,
            ):
                if existing_label.name not in unhandled_symbols:
                    raise SymbolExistsError(
                        f"Multiple LabeledAddress resources with name {existing_label.name}!"
                    )
                existing_label.resource.add_view(unhandled_symbols.pop(existing_label.name))
        except NotFoundError:
            LOGGER.debug("No existing LabeledAddresses found for the provided symbols, moving on")

        # Overwrite existing Addressables with new LinkableSymbols
        # Only do this for LinkableSymbols which did not override existing LabeledAddress above
        unhandled_vaddrs = {symbol.virtual_address: symbol for symbol in unhandled_symbols.values()}

        filter_for_label_by_vaddr = ResourceFilter(
            tags=(Addressable,),
            attribute_filters=(
                ResourceAttributeValuesFilter(
                    Addressable.VirtualAddress, tuple(unhandled_vaddrs.keys())
                ),
            ),
        )

        try:
            for existing_addressable in await resource.get_descendants_as_view(
                Addressable,
                r_filter=filter_for_label_by_vaddr,
            ):
                symbol = unhandled_vaddrs[existing_addressable.virtual_address]
                existing_addressable.resource.add_view(symbol)
                unhandled_symbols.pop(symbol.name)
        except NotFoundError:
            LOGGER.debug("No existing Addressable found for the provided symbols, moving on")

        # For any new LinkableSymbols that we could not add to existing resources, create a new
        # resource
        for symbol in unhandled_symbols.values():
            await resource.create_child_from_view(symbol)



import asyncio
import os
import pathlib
import sys
import tempfile
from subprocess import CalledProcessError
from dataclasses import dataclass

from ofrak.core.filesystem import File, Folder

from ofrak.component.packer import Packer

from ofrak.resource import Resource

from ofrak.component.unpacker import Unpacker
from ofrak.component.identifier import Identifier

from ofrak.model.component_model import ComponentConfig, ComponentExternalTool
from ofrak.core.zip import ZipArchive, UNZIP_TOOL
from ofrak.core.binary import GenericBinary
from ofrak.core.magic import Magic, MagicMimeIdentifier
from ofrak_type.range import Range


APKTOOL = ComponentExternalTool("apktool", "https://ibotpeaches.github.io/Apktool/", "--help")
JAVA = ComponentExternalTool(
    "java",
    "https://openjdk.org/projects/jdk/11/",
    "--help",
    apt_package="openjdk-11-jdk",
    brew_package="openjdk@11",
)


class _UberApkSignerTool(ComponentExternalTool):
    if sys.platform.startswith("win32"):
        # Windows: look in Program Files (x86)
        JAR_PATH = os.path.join(
            "C:", "Program Files (x86)", "uber-apk-signer", "uber-apk-signer.jar"
        )
    elif sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
        # Linux, Mac OSX: look in usr/local/bin
        JAR_PATH = os.path.join(os.path.sep, "usr", "local", "bin", "uber-apk-signer.jar")
    else:
        # All other platforms: look in home dir
        JAR_PATH = os.path.join(pathlib.Path.home(), "uber-apk-signer.jar")

    def __init__(self):
        super().__init__(
            _UberApkSignerTool.JAR_PATH,
            "https://github.com/patrickfav/uber-apk-signer",
            install_check_arg="",
        )

    async def is_tool_installed(self) -> bool:
        if not os.path.exists(_UberApkSignerTool.JAR_PATH):
            return False

        try:
            cmd = [
                "java",
                "-jar",
                _UberApkSignerTool.JAR_PATH,
                "--help",
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            returncode = await proc.wait()
        except FileNotFoundError:
            return False

        return 0 == returncode


UBER_APK_SIGNER = _UberApkSignerTool()


class Apk(ZipArchive):
    pass


class ApkUnpacker(Unpacker[None]):
    """
    Decode Android APK files.

    This unpacker is a wrapper for `apktool`. See <https://ibotpeaches.github.io/Apktool/>.
    """

    targets = (Apk,)
    children = (File, Folder)
    external_dependencies = (APKTOOL,)

    async def unpack(self, resource: Resource, config=None):
        """
        Decode Android APK files.

        :param resource:
        :param config:
        """
        apk = await resource.view_as(Apk)
        data = await resource.get_data()
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_file.write(data)
            temp_file.flush()
            with tempfile.TemporaryDirectory() as temp_flush_dir:
                cmd = [
                    "apktool",
                    "decode",
                    "--output",
                    temp_flush_dir,
                    "--force",
                    temp_file.name,
                ]
                proc = await asyncio.create_subprocess_exec(
                    *cmd,
                )
                returncode = await proc.wait()
                if proc.returncode:
                    raise CalledProcessError(returncode=returncode, cmd=cmd)
                await apk.initialize_from_disk(temp_flush_dir)


@dataclass
class ApkPackerConfig(ComponentConfig):
    sign_apk: bool


class ApkPacker(Packer[ApkPackerConfig]):
    """
    Pack decoded APK resources into an APK.

    This unpacker is a wrapper for two tools:

    - `apktool` repacks the APK resources. See <https://ibotpeaches.github.io/Apktool/>.
    - `uber-apk-signer` signs the packed APK file. See
    <https://github.com/patrickfav/uber-apk-signer>.

    Another helpful overview of the process: <https://github.com/vaibhavpandeyvpz/apkstudio>.
    """

    targets = (Apk,)
    external_dependencies = (APKTOOL, JAVA, UBER_APK_SIGNER)

    async def pack(
        self, resource: Resource, config: ApkPackerConfig = ApkPackerConfig(sign_apk=True)
    ):
        """
        Pack disassembled APK resources into an APK.

        :param resource:
        :param config:
        """
        apk = await resource.view_as(Apk)
        temp_flush_dir = await apk.flush_to_disk()
        apk_suffix = ".apk"
        with tempfile.NamedTemporaryFile(suffix=apk_suffix) as temp_apk:
            apk_cmd = [
                "apktool",
                "build",
                "--force-all",
                temp_flush_dir,
                "--output",
                temp_apk.name,
            ]
            apk_proc = await asyncio.create_subprocess_exec(
                *apk_cmd,
            )
            apk_returncode = await apk_proc.wait()
            if apk_proc.returncode:
                raise CalledProcessError(returncode=apk_returncode, cmd=apk_cmd)
            if not config.sign_apk:
                # Close the file handle and reopen, to avoid observed situations where temp.read()
                # was not returning data
                with open(temp_apk.name, "rb") as file_handle:
                    new_data = file_handle.read()
            else:
                with tempfile.TemporaryDirectory() as signed_apk_temp_dir:
                    java_cmd = [
                        "java",
                        "-jar",
                        _UberApkSignerTool.JAR_PATH,
                        "--apks",
                        temp_apk.name,
                        "--out",
                        signed_apk_temp_dir,
                        "--allowResign",
                    ]
                    java_proc = await asyncio.create_subprocess_exec(
                        *java_cmd,
                    )
                    java_returncode = await java_proc.wait()
                    if java_proc.returncode:
                        raise CalledProcessError(returncode=java_returncode, cmd=java_cmd)
                    signed_apk_filename = (
                        os.path.basename(temp_apk.name)[: -len(apk_suffix)]
                        + "-aligned-debugSigned.apk"
                    )
                    signed_file_name = os.path.join(
                        signed_apk_temp_dir,
                        signed_apk_filename,
                    )
                    with open(signed_file_name, "rb") as file_handle:
                        new_data = file_handle.read()
            assert len(new_data) != 0
            resource.queue_patch(Range(0, await resource.get_data_length()), new_data)


class ApkIdentifier(Identifier):
    targets = (File, GenericBinary)
    external_dependencies = (UNZIP_TOOL,)

    async def identify(self, resource: Resource, config=None) -> None:
        await resource.run(MagicMimeIdentifier)
        magic = resource.get_attributes(Magic)
        if magic is not None and magic.mime in ["application/java-archive", "application/zip"]:
            with tempfile.NamedTemporaryFile(suffix=".zip") as temp_file:
                temp_file.write(await resource.get_data())
                temp_file.flush()
                unzip_cmd = [
                    "unzip",
                    "-l",
                    temp_file.name,
                ]
                unzip_proc = await asyncio.create_subprocess_exec(
                    *unzip_cmd,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
                stdout, stderr = await unzip_proc.communicate()
                if unzip_proc.returncode:
                    raise CalledProcessError(returncode=unzip_proc.returncode, cmd=unzip_cmd)

                if b"androidmanifest.xml" in stdout.lower():
                    resource.add_tag(Apk)



from dataclasses import dataclass

from ofrak.model.resource_model import index

from ofrak.core.addressable import Addressable


@dataclass
class LabeledAddress(Addressable):
    name: str

    @index
    def Label(self) -> str:
        return self.name
